# -*- coding: utf-8 -*-
"""
Created on Wed Nov  3 17:39:15 2021

@author: ScottStation
"""
import time
import json
from datetime import datetime, timedelta
from .qetype import qetype
from qesdk import is_valid_trade_time
from .qeredisdb import getDBPositionData,getDBOrderData,getDBTradeData,saveTradeDataToDB,saveOrderDataToDB
from .qeredisdb import saveNewAccountToDB, loadSimuAccounts,removeDBSimuAccounts,getDBAccountData
from .qeredisdb import getSimuInitCap,updateInitCap,saveAccountDataToDB,savePositionDataToDB
from .qelogger import logger
from .qestatistics import g_stat
from .qeglobal import instSetts
from .qeriskctl import riskControl
import numpy as np
import collections
class qeDataSlide(collections.UserDict):
    def __missing__(self, key):
        key = str(key)
        if not key in self.data:
            return {"current":0,"presett":0,"preclose":0}    
        return self.data[key]
 
    def __contains__(self, key):
        return str(key) in self.data
 
    def __setitem__(self, key, item):
        self.data[str(key)] = item
    
    def __getitem__(self, key):
        key = str(key)
        if not key in self.data:
            return {"current":0,"presett":0,"preclose":0}    
        return self.data[key]


#tstrats = {}
#instSetts = {}
g_dataSlide = qeDataSlide()
feesmult = 1.0

from threading import Timer

def getMixPrice(oldp, oldv, newp, newv):
    return(oldp*oldv + newp*newv)/(oldv+newv)

def getCurTradingDay():
    now = datetime.now()
    wday = now.weekday()
    if now.hour >= 19:
        if wday < 5:
            now += timedelta(days=1)
        else:    
            now += timedelta(days=3)
            
    elif now.hour < 8 and wday == 5:
        now += timedelta(days=2)
    return int(now.strftime('%Y%m%d'))            


class accountInfo():
    def __init__(self):
        self.balance = 10000000
        self.avail = self.balance
        self.margin = 0
        self.maxmarg = 0
        self.frozenMarg = 0
        self.pendVolumes = {}
        self.frozenVol = {}
        self.position = {}
        self.trades = {}
        self.orders = {}
        
        #self.g_order_id = int('3'+datetime.now().strftime('%Y%m%d')[2:])*100000
        self.g_trade_id = int('3'+datetime.now().strftime('%Y%m%d')[2:])*100000
        self.closeProf = 0
        self.posProf = 0
        self.daypnl = 0
        self.totalpnl = 0
        self.dayfees = 0
        self.totalfees = 0
        self.winamount = 0
        self.lossamount = 0
        self.wincount = 0
        self.losscount = 0
        self.turnover = 0
        self.current_timedigit = 0
        self.tradingDay = ""
        self.user = "unknown"
        self.token = ""
        self.curtime = datetime.now()
        self.riskctl = riskControl(self.riskctlCall)
        self.loadReady = False
        self.loadPosition = True
        self.stratCross = {}
        self.callback = None
        
   
    def crossday(self):
        logger.info('simuaccount.crossday')
        self.dayfees = 0
        self.daypnl = 0
        self.trades = {}
        self.orders = {}
        #self.g_order_id = int('3'+datetime.now().strftime('%Y%m%d')[2:])*100000
        self.g_trade_id = int('3'+datetime.now().strftime('%Y%m%d')[2:])*100000
        self.pendVolumes = {}
        self.frozenMarg = 0
        self.winamount = 0
        self.lossamount = 0
        self.wincount = 0
        self.maxmarg = 0
        self.losscount = 0
        self.turnover = 0
        self.riskctl.crossDay()
        for inst in self.position.keys():
            self.position[inst]['long']['yesvol'] = self.position[inst]['long']['volume']
            self.position[inst]['short']['yesvol'] = self.position[inst]['short']['volume']
        self.checkforceclose()
        
    def setLoadReady(self):
        self.loadReady = True
        
    def riskctlCall(self):
        instPrices = {}
        for oid in self.orders:
            order = self.orders[oid]
            if order['leftvol'] > 0 and order['ordertype'] == 'limit':
                field = 'longhigh' if order['direction'] > 0 else 'shortlow'
                instid = order['instid']
                price = order['price']
                if not instid in instPrices:
                    instPrices[instid] = {field:price}
                elif not field in  instPrices[instid]:
                    instPrices[instid][field] = price
                else:
                    if field == 'longhigh' and price > instPrices[instid][field]:
                        instPrices[instid][field] = price
                    elif field == 'shortlow' and price < instPrices[instid][field]:
                        instPrices[instid][field] = price     
        return instPrices                
    
    def getPosProf(self):
        global g_dataSlide
        posProf = 0
        for inst in self.position.keys():
            if inst in g_dataSlide:# and inst in instSetts:
                current = g_dataSlide[inst]['current']
                volMult = instSetts[inst]['volmult']
                if self.position[inst]['long']['volume'] > 0:
                    posProf += volMult * (current - self.position[inst]['long']['poscost']) * \
                               self.position[inst]['long']['volume']
                if self.position[inst]['short']['volume'] > 0:
                    posProf += volMult * (self.position[inst]['short']['poscost'] - current) * \
                               self.position[inst]['short']['volume']
        return posProf                   
    
    def getPositionVol(self,instid):
        totalL = self.getPosition(instid, 'long', 'volume')
        totalS = self.getPosition(instid, 'short', 'volume')
        yesL = self.getPosition(instid, 'long','yesvol')
        yesS = self.getPosition(instid, 'short', 'yesvol')
        return totalL - yesL , yesL, totalS - yesS, yesS    
    
    def updateMargin(self):
        global g_dataSlide
        marg = 0
        for instid in self.position.keys():
            #current = g_dataSlide[inst]['current']
            tL, yL, tS, yS = self.getPositionVol(instid)
            margL = self.getMargin(self.getPosition(instid, 'long', 'poscost'), True, tL, instid) 
            if yL > 0:
                margL += self.getMargin(g_dataSlide[instid]['presett'], True, yL, instid)
            margS = self.getMargin(self.getPosition(instid, 'short', 'poscost'),False, tS, instid)
            if yS > 0:
                margS += self.getMargin(g_dataSlide[instid]['presett'], False, yS, instid)
            marg += margL + margS
            #marg += self.getMargin(self.position[inst]['long']['poscost'], True, self.position[inst]['long']['volume'], inst)
            #marg += self.getMargin(self.position[inst]['short']['poscost'], False, self.position[inst]['short']['volume'], inst)
        self.margin = marg
        self.maxmarg = max(marg, self.maxmarg)
        #print("account margin",self.margin, self.position)
        
    def updateFrozenMarg(self):
        global g_dataSlide
        fm = 0
        for oid in self.orders:
            order = self.orders[oid]
            if order['leftvol'] >0 and order['action'] == 'open':
                inst = order['instid']
                odir = order['direction']
                if order['ordertype'] == 'market':
                    current = g_dataSlide[inst]['current'] if inst in g_dataSlide else 0
                else:
                    current = order['price']
                if current > 0:
                    fm += self.getMargin(current, odir > 0, order['leftvol'], inst)
        self.frozenMarg = fm        
    
    #def updateFrozenMarg(self):
    #    global g_dataSlide
    #    fm = 0
    #    for inst in self.pendVolumes.keys():
    #        #current = g_dataSlide[inst]['current']
    #        fm += self.getMargin(self.pendVolumes[inst]['long']['poscost'], True, self.pendVolumes[inst]['long']['volume'], inst)
    #        fm += self.getMargin(self.pendVolumes[inst]['short']['poscost'], False, self.pendVolumes[inst]['short']['volume'], inst)
    #    self.frozenMarg = fm
    def updateStratPostion(self, pos):
        def mixPosition(cost1, vol1, cost2, vol2):
            cost  =  (cost1*vol1 + cost2*vol2)/(vol1+vol2) if vol1+vol2 > 0 else 0
            vol = vol1 + vol2
            return {'volume': vol, 'poscost': cost, 'yesvol': 0}
            
            
        for inst in pos:
            if inst in self.position:
                for dirstr in pos[inst]:
                    if dirstr in self.position[inst]:
                        self.position[inst][dirstr] = mixPosition( self.position[inst][dirstr]['poscost'], self.position[inst][dirstr]['volume'], pos[inst][dirstr]['poscost'], pos[inst][dirstr]['volume'] )
                    else:
                        self.position[inst][dirstr] = pos[inst][dirstr]
                    
            else:
                self.position[inst] = pos[inst]
        
        
        
    def updateCapital(self, closeProf, fee):
        lastPosProf = self.posProf
        self.posProf = self.getPosProf()
        #print(type(closeProf), type(fee), type(lastPosProf), type(self.posProf))
        self.balance += closeProf - fee -lastPosProf + self.posProf
        self.avail = self.balance -self.margin - self.frozenMarg
    
    def getMargin(self, current, bLong, vol, instid):
        marginrate = instSetts[instid]['marglong'] if bLong else instSetts[instid]['margshort']
        return current * marginrate / 100 * vol * instSetts[instid]['volmult']

    def getCommission(self,  action, price, vol, closetype,instid):
        #global instSetts
        
        if action == 'open':

            return instSetts[instid]['openfee'] * vol + instSetts[instid]['openfeerate'] * vol * price * \
                   instSetts[instid]['volmult']
        else:
            rate = instSetts[instid]['closetodayrate'] if closetype == 'closetoday' else 1
            return  rate * (instSetts[instid]['closefee'] * vol + \
                                                          instSetts[instid]['closefeerate'] * vol * price * instSetts[instid]['volmult'])

    def readAndSetDB(self):
        import math
        try:
            #load from db ,if not exist, write default value
            bal =  getSimuInitCap(self.user)
            if bal is None or math.isnan(float(bal)):
                updateInitCap(self.user, self.balance)
            else:
#                 logger.info(f'load balance:{self.balance}')
                self.balance = float(bal)
        except Exception as e:
            logger.error(f"account.readAndSetDB error {e}",exc_info=True ) 
          
            
    
    def saveToDB(self):
        d = {}
        # d['tradingday'] = self.tradingDay 
        # d['timedigit'] = self.current_timedigit 
#         d['initcapital'] = account.initCapital
#         d['curpnl'] = account.curpnl
        d['posProf'] = str(round(self.posProf,3))
        d['frozenMarg'] = str(round(self.frozenMarg,3))
        d['marg'] = str(round(self.margin,3))
        d['closeProf'] = str(round(self.closeProf,3))
        d['balance'] = str(round(self.balance,3))
        d['avail'] = str(round(self.avail,3))
        d['totalpnl'] = str(round(self.totalpnl,3))
        d['totalfee'] = str(round(self.totalfees,3))
        d['daypnl'] = str(round(self.daypnl,3))
        d['dayfee'] = str(round(self.dayfees,3))
        d['tradingDay'] = str(self.tradingDay)
        d['winamount'] = str(round(self.winamount, 3))
        d['lossamount'] = str(round(self.lossamount, 3))
        d['wincount'] = str(round(self.wincount, 3))
        d['losscount'] = str(round(self.losscount, 3))
        d['turnover'] = str(round(self.turnover, 3))
        d['maxmarg'] = str(round(self.maxmarg, 3))
        
        saveAccountDataToDB(self.user, self.token,  d)
        savePositionDataToDB(self.user, self.token, self.position)
        
        g_stat.updateData(self.balance, self.daypnl, self.dayfees, self.margin, self.turnover, self.tradingDay,\
                          self.winamount, self.lossamount, self.wincount, self.losscount,self.maxmarg)
        #updateInitCap(self.user, self.balance)
        #logger.info(f'saveToDB {self.balance}')
    
    def getPosition(self, instid, dirstr,feild='volume'):
         if instid in self.position:
            if dirstr in self.position[instid]:
                if feild in self.position[instid][dirstr]:
                    return self.position[instid][dirstr][feild]
         return 0
       
    '''
    def getFrozenVol(self, instid, dirstr):
        if instid in self.frozenVol:
            if dirstr in self.frozenVol[instid]:
                return self.frozenVol[instid][dirstr]
        return 0
    '''
    
    def calcFrozenVol(self, instid, direction):
        fvol = 0
        for oid in self.orders:
            order = self.orders[oid]
            if order['instid'] == instid and order['leftvol'] > 0 and order['action']=='close' and order['direction'] == direction:
                fvol += order['leftvol']
        return fvol
        
    def makeOrder(self, action, direction, price, vol, instid):
        margin = self.getMargin(price, direction>0, vol, instid)
        if action == 'open' and self.avail <= margin:
            return -1
        elif action == 'open':
            pass
            '''
            if direction > 0:
                if instid in self.pendVolumes:
                    self.pendVolumes[instid]['long']['poscost'] = getMixPrice(self.pendVolumes[instid]['long']['poscost'],self.pendVolumes[instid]['long']['volume'],price,vol)
                    self.pendVolumes[instid]['long']['volume'] += vol
                else:
                    self.pendVolumes[instid]={'long':{'volume':vol,'poscost':price}, "short":{'volume':0,"poscost":0}}
            else:
                if instid in self.pendVolumes:
                    self.pendVolumes[instid]['short']['poscost'] = getMixPrice(self.pendVolumes[instid]['short']['poscost'],self.pendVolumes[instid]['short']['volume'],price,vol)
                    self.pendVolumes[instid]['short']['volume'] += vol
                else:
                    self.pendVolumes[instid]={'long':{'volume':0,"poscost":0}, "short":{'volume':vol,'poscost':price}}
            '''
        elif action == 'close':
            dirstr = 'short' if direction > 0 else 'long'
            pend = self.calcFrozenVol(instid, direction)
            #fvol = self.getFrozenVol(instid, dirstr)
            pos = self.getPosition(instid, dirstr)
            leftvol = pos - pend
            if vol > leftvol:
                print(f'Position is not enough: vol{vol}, pos {pos}, pend {pend}, left {leftvol}')
                logger.warning(f'Position is not enough: vol{vol},  pos {pos}, pend {pend}, left {leftvol}')
                return -2
            #else:
            #    if not instid in self.frozenVol:
            #        self.frozenVol[instid] = {"long":0, "short":0}
            #    self.frozenVol[instid][dirstr] += vol
                
              
                
                
        if instid[-3:]=='CCF':
            self.dayfees += 1
            self.totalfees += 1
        self.updateFrozenMarg()  
        self.avail =self.balance - self.margin - self.frozenMarg
        self.saveToDB()
        return 0    
        
        
    def cancelOrder(self, vol, direction, instid, action, saveDB=True):
        if action =='open':
            dirstr = 'long' if direction > 0 else 'short'
            #self.pendVolumes[instid][dirstr]['volume'] = max(self.pendVolumes[instid][dirstr]['volume']-vol,0)
            #if self.pendVolumes[instid][dirstr]['volume'] == 0:
            #    self.pendVolumes[instid][dirstr]['poscost'] = 0
                
            self.updateFrozenMarg()
            self.avail =self.balance - self.margin - self.frozenMarg

        #elif action =='close':
        #    dirstr = 'short' if direction > 0 else 'long'
            #self.frozenVol[instid][dirstr] = max(self.frozenVol[instid][dirstr] - vol, 0)
         
        
        if instid[-3:]=='CCF':
            self.dayfees += 1
            self.totalfees += 1
        if saveDB:
            self.saveToDB()    
        
        
    def  onTrade(self, instid, action, blong,  price, vol, closetype,feesmult=1.0):
        global g_dataSlide   #,  tstrats
        
        def updatePos(poscost, posvol, price, vol):
            if vol + posvol > 0:
                poscost = (poscost*posvol + price*vol)/(vol+posvol)
            else:
                poscost = 0
            posvol = vol + posvol
            return poscost, posvol
        try:
            
            dirstr = 'long' if blong else 'short'
            closeProf = 0
            if action=='open':
                ## update position
                if not instid in self.position:
                    if blong:
                        self.position[instid] = {'long':{'poscost':price,'volume':vol, 'yesvol':0}, \
                                                 'short':{'poscost':0,'volume':0, 'yesvol': 0}}
                    else:
                        self.position[instid] = {'short':{'poscost':price,'volume':vol, 'yesvol':0}, \
                                                 'long':{'poscost':0,'volume':0, 'yesvol':0}}
                else:

                    self.position[instid][dirstr]['poscost'],self.position[instid][dirstr]['volume'] = updatePos(self.position[instid][dirstr]['poscost'],self.position[instid][dirstr]['volume'], price, vol)
                #print('OnTrade', price, vol, self.position)

            elif action=='close':   
                volMult = instSetts[instid]['volmult']
                if dirstr == 'short':
                    closeProf = volMult * vol * (self.position[instid][dirstr]['poscost'] - price)
                else:
                    closeProf = volMult * vol * (price - self.position[instid][dirstr]['poscost'])

                self.daypnl += closeProf
                self.totalpnl += closeProf
                self.position[instid][dirstr]['volume'] -= vol
                #self.frozenVol[instid][dirstr] = max(self.frozenVol[instid][dirstr]-vol, 0)
                if closetype == 'closeyesterday':
                    self.position[instid][dirstr]['yesvol'] -= vol
                if self.position[instid][dirstr]['volume'] <= 0:
                    self.position[instid][dirstr] = {'poscost':0,'volume':0,'yesvol':0}
                    
                tfee = feesmult *self.getCommission("open",self.position[instid][dirstr]['poscost'] , vol, closetype, instid)
                tfee += feesmult * self.getCommission(action,price,vol, closetype,instid)
                tpnl = closeProf -tfee
                if tpnl > 0:
                    simuaccount.wincount += 1
                    simuaccount.winamount += tpnl
                elif tpnl < 0:
                    simuaccount.losscount += 1
                    simuaccount.lossamount += tpnl
                   
                
            ##update pendVolume
            '''
            if action == 'open':
                self.pendVolumes[instid][dirstr]['volume'] = max(self.pendVolumes[instid][dirstr]['volume'] -  vol, 0)
                if self.pendVolumes[instid][dirstr]['volume'] == 0:
                    self.pendVolumes[instid][dirstr] = {'poscost':0,'volume':0}
            '''
            
            self.turnover += price * vol * instSetts[instid]['volmult']
            self.updateFrozenMarg()
            self.updateMargin()
            fee = feesmult * self.getCommission(action,price,vol, closetype,instid)
            self.totalfees += fee
            self.dayfees += fee
            

            self.updateCapital(closeProf, fee)
            self.saveToDB()
        except Exception as e:
            logger.error(f"account.onTrade error {e}",exc_info=True ) 
    
        
    def autoAction(self, instid, blong, vol):
        checkdir = 'short' if blong else 'long'
        if self.position[instid][checkdir]['volume'] > 0:
            tvol = min(self.position[instid][checkdir]['volume'], vol)
            return 'close', tvol
        else:
            return 'open', vol
        
        
    
    def checkCloseToday(self, instid, blong, closetype, vol):
        dirstr = 'long' if blong else 'short'
        yesvol = self.getPosition(instid,dirstr,'yesvol')
        dayvol = self.getPosition(instid,dirstr,'volume') - self.getPosition(instid,dirstr,'yesvol')
        #print(closetype,self.position, instid, dirstr,  yesvol, dayvol, vol)
        if closetype == 'closeyesterday':
            return min(yesvol,vol), closetype
        elif closetype == 'closetoday':
            return min(dayvol,vol),closetype
        elif yesvol > 0:
            return min(yesvol, vol), 'closeyesterday'
        else:
            return min(dayvol, vol), 'closetoday'
    
    #def noticeCrossDay(self, name):
    #    if name in self.stratCross:
    #        self.stratCross[name] = 1
    #    if sum(self.stratCross.values()) == len(self.stratCross):
    #        self.checkForceClose()
    #        for name in self.stratCross:
    #            self.stratCross = 0
    def getFCVol(self, isToday, price, bLong, cost, presett, instid):
        volmult = instSetts[instid]['volmult']
        closetype = 'closetoday' if isToday else 'closeyesterday'
        factor1 = self.getMargin(price, bLong, 1, instid)
        cost = presett if not isToday else cost
        priceDiff = price - cost if bLong else cost - price 
        factor2 = priceDiff * volmult - self.getCommission('close',price, 1,closetype,instid)
        return factor1, factor2
        #return np.ceil((marg - balance) / (factor1 + factor2) )
                
    def calcForceClose(self, instid, marg, balance):
        global g_dataSlide
        A = marg
        B = balance
        tL, yL, tS, yS = self.getPositionVol(instid)
        ps = g_dataSlide[instid]['presett']
        costL, costS = self.getPosition(instid, 'long', 'poscost'), self.getPosition(instid, 'short', 'poscost'), 
        price = g_dataSlide[instid]['current']
        finished = False
        fclist = [0,0,0,0]
        if yL > 0:
            f1, f2 = self.getFCVol(False, price, True, costL, ps, instid)
            fcvol = min(np.ceil((A-B) /(f1+f2)),yL)
            A -= f1 * fcvol
            B += f2 * fcvol
            fclist[0] = fcvol
            finished = A < B
        if not finished and yS > 0:
            f1, f2 = self.getFCVol(False, price, False, costS, ps, instid)
            fcvol = min(np.ceil((A-B) /(f1+f2)),yS)
            A -= f1 * fcvol
            B += f2 * fcvol
            fclist[1] = fcvol
            finished = A < B
        if not finished and tL > 0:
            f1, f2 = self.getFCVol(True, price, True, costL, ps, instid)
            fcvol = min(np.ceil((A-B) /(f1+f2)),tL)
            A -= f1 * fcvol
            B += f2 * fcvol
            fclist[2] = fcvol
            finished = A < B
        if not finished and tS > 0:
            f1, f2 = self.getFCVol(True, price, False, costS, ps, instid)
            fcvol = min(np.ceil((A-B) /(f1+f2)),tS)
            A -= f1 * fcvol
            B += f2 * fcvol
            fclist[3] = fcvol
            finished = A < B
        return finished, A, B, fclist
          
    
    
    def checkforceclose(self):
        if self.balance <  self.margin:
            ## touch force close
            ## get max marg instid
            A = self.margin
            B = self.balance
            fcDict = {}
            for instid in self.position:
                finished, A, B, fclist = self.calcForceClose(instid, A, B)
                fcDict[instid] = fclist
                if finished:
                    break
            print('权益低于保证金，触发平台强平操作')
            
            if self.callback:
                self.callback(fcDict)
            #force_close(self, fcDict)
    
simuaccount = accountInfo()           

def createSimuAccount(user, initCap=10000000):
    '''
    

    Parameters
    ----------
    user : str
        username.
    initCap : float, optional
        initial capital. The default is 10000000.

    Returns
    -------
    token : str
        token of new simulation account.

    '''
    import uuid
    token = str(uuid.uuid1()).replace('-', '')
    
    simuacc = accountInfo()
    simuacc.user = user
    simuacc.token = token
    simuacc.balance = initCap
    saveNewAccountToDB(user, token)
    simuacc.saveToDB()
    
    return token
    
def removeSimuAccountData(user, tokens=None, excepts=None):
    tlist = []
    if tokens:
         tlist = tokens
    else:
         tlist = loadSimuAccounts(user)
         if excepts:
             tlist = [t for t in tlist if not t in excepts]
    return removeDBSimuAccounts(user, tlist)         


def listSimuAccounts(user):
    ret = loadSimuAccounts(user)
    if ret:
        return ret
    else:
        return []             

class QEsimtrader(object):

    def __init__(self):
        
        self.tqueue = None
        self.strats = None
        self.curday = ''
        self.instVolume={}
        #self.g_orders = {}
        #self.g_trades = {}
        #self.g_order_id = int('3'+datetime.now().strftime('%Y%m%d')[2:])*100000
        #self.g_trade_id = int('3'+datetime.now().strftime('%Y%m%d')[2:])*100000
        
        #self.account = accountInfo()
        #simuaccount.balance=0
        
        #self.g_benchmark = self.g_order_id
        self.lasttime = 0
        self.user = 'unknown'
        self.Timer_interval = 60
        self.mapTable = {}     
    
    def callTimer(self): 
#         global tqueue
        timer = Timer(self.Timer_interval,self.callTimer)
        d = {}
        d['type'] = qetype.KEY_TIMER_SIMU
#         logger.info('timer '+str(datetime.now()))
        
        if self.tqueue:
            self.tqueue.put(d)
        else:
            logger.info('Timer is pending')
        timer.start()
        return
    def heartBeat(self,realTrade):
        pass
        '''
        d = {}
        d['time'] = datetime.now().strftime('%Y%m%d%H%M%S')
        if realTrade:
            d['type'] = 'real'
        else:
            d['type'] = 'simu'
#         print('heartbeat '+str(d['time']))
        saveProcessToDB(self.user,d)
        '''
    
    
    def loadFromDB(self, tradingDay):
        #tradingDay = getCurTradingDay()
        acc = getDBAccountData(self.user, self.token)
        simuaccount.user = self.user
        simuaccount.token = self.token
        logger.info(f'simuaccount.loadFromDB {tradingDay}')
        if acc:
            d = json.loads(acc)
            simuaccount.posProf = float(d['posProf'])
            simuaccount.frozenMarg = 0
            simuaccount.margin =   float(d['marg']) 
            simuaccount.closeProf =  float(d['closeProf'])
            simuaccount.balance = float(d['balance'])
            simuaccount.avail =  float(d['avail'])
            simuaccount.totalpnl = float(d['totalpnl'])
            simuaccount.totalfees = float(d['totalfee']) 
            simuaccount.daypnl = float(d['daypnl']) 
            simuaccount.dayfees = float(d['dayfee']) 
            simuaccount.winamount = float(d['winamount'])
            simuaccount.lossamount = float(d['lossamount'])
            simuaccount.wincount = float(d['wincount'])
            simuaccount.losscount = float(d['losscount'])
            simuaccount.turnover = float(d['turnover'])
            simuaccount.maxmarg = float(d['maxmarg'])
            
            #print(tradingDay, d)
            if d['tradingDay'] != '' and int(d['tradingDay']) < int(tradingDay):
                simuaccount.crossday()
            simuaccount.tradingDay = tradingDay
        else:
            logger.warn(f"读取模拟账户数据失败, user:{self.user},token:{self.token}.")
        
        if simuaccount.loadPosition:     
            pos = getDBPositionData(self.user, self.token)
            if pos:
                simuaccount.position = json.loads(pos)
        #else:
        #    logger.warn(f"读取模拟账户持仓数据失败, user:{self.user},token:{self.token}.")
        
        orders = getDBOrderData(self.user, self.token, tradingDay)
        for oid in orders:
            content = eval(orders[oid])
            simuaccount.orders[oid] = content
        
        trades = getDBTradeData(self.user,self.token, tradingDay)
        for tid in trades:
            content = eval(trades[tid])
            simuaccount.trades[tid] = content
        
        #print("load",simuaccount.position, simuaccount.margin, simuaccount.balance)
        ###Deal with unfinished orders
        
        for oid in simuaccount.orders:
            #print(oid)
            #print(simuaccount.orders[oid])
            #print(type(simuaccount.orders[oid]))
            if simuaccount.orders[oid]['leftvol'] > 0:
                simuaccount.orders[oid]['status'] = qetype.KEY_STATUS_PTPC if simuaccount.orders[oid]['tradevol'] > 0 else qetype.KEY_STATUS_CANCEL
                simuaccount.orders[oid]['cancelvol'] += simuaccount.orders[oid]['leftvol']
                simuaccount.orders[oid]['leftvol'] = 0
        g_stat.loadFromDBSimu(int(tradingDay))
        simuaccount.setLoadReady()
    
    def sendForceCloseOrder(self, instid, direction, price, volume, ordertype="limit", action="open", closetype='auto'):
        now = datetime.now()
        orderid = g_stat.getNewOrderID()
        temporder = {'instid': instid,
                     'price': price,
                     'direction': direction,
                     'ordertype': ordertype,
                     'closetype': closetype,
                     'volume': volume,
                     'leftvol': volume,
                     'tradevol': 0,
                     'cancelvol': 0,
                     'pendvol': 0,
                     'status': 'committed',
                     'errorid': 0,
                     'errormsg': '',
                     'action': action,
                     'timecond': 'GFD',
                     'autoremake':[0,0,0,0,now,'wait'],
                     'autocancel': [0,0,0,0,now,'wait'],
                     'father': 0,
                     'accid': 0}

        temporder['type'] = qetype.KEY_SEND_ORDER
        temporder['stratName'] = 'force_close'
        temporder['orderid'] = orderid
        
        self.tqueue.put(temporder)
    
    
    
    def accountCallback(self, fcDict):
        global g_dataSlide
        for inst in fcDict:
            current = g_dataSlide[inst]['current']
            fclist = fcDict[inst]
            if fclist[0] > 0:
                self.sendForceCloseOrder(inst, -1, current,fclist[0],'market', 'close', 'closeyesterday')
            if fclist[1] > 0:
                self.sendForceCloseOrder(inst, 1, current,fclist[1],'market', 'close', 'closeyesterday')
            if fclist[2] > 0:    
                self.sendForceCloseOrder(inst, -1, current,fclist[2],'market', 'close', 'closetoday')
            if fclist[3] > 0:
                self.sendForceCloseOrder(inst, 1, current,fclist[3],'market', 'close', 'closetoday')
        
    
    
    def simTraderProcess(self):      
        logger.info(u"模拟器启动")
        print(u"模拟器启动")
        #simuaccount.feesmult = feesMult
        
        simuaccount.user = self.user
        #simuaccount.stratCross = {key : 0 for key in self.strats}
        simuaccount.callback = self.accountCallback
        #simuaccount.readAndSetDB()
        self.callTimer()
    
        if self.tqueue:
            #print('simtrader tqueue is ready')
            while True:
                
                while not self.tqueue.empty():
                    try:
                        d = self.tqueue.get(block = True, timeout = 1)
                        self.process(d)
                        
                    except Exception as e:
                        logger.error(f"simtrader tqueue error {e}",exc_info=True ) 
                          
                time.sleep(0.001)
            

    def process(self,d):
       
        if d['type'] == qetype.KEY_SEND_ORDER:
            self.sendOrder(d)
        elif d['type'] == qetype.KEY_CANCEL_ORDER:
            self.cancelOrder(d)
        elif d['type'] == qetype.KEY_MARKET_DATA:
            #print('trader mdata')
            self.update(d)
        elif d['type'] == qetype.KEY_TIMER_SIMU:
            #orderlen = len(simuaccount.orders)
            #logger.debug(f'simtrader.onTimer {orderlen}')
            simuaccount.updateCapital(0,0)
            simuaccount.saveToDB()
            
            #self.heartBeat(False)

    def callback(self,d):
        if self.strats:
            if self.strats['async']:
	            if d['stratName'] != 'force_close':
	                self.strats['queue'].put(d)
            else:
	            stratSett = self.strats.get(d['stratName'],None)
	            if stratSett:
	                stratSett['queue'].put(d)
	            elif d['stratName'] != 'force_close':
	                logger.error('callback '+str(d['stratName'])+' is not found')

    def crossday(self):
        #self.g_order_id = self.g_benchmark
        logger.info('simTrader.crossday')
        self.instVolume = {}
        self.mapTable = {}     
        #self.g_order_id = int('3'+datetime.now().strftime('%Y%m%d')[2:])*100000
        #self.g_trade_id = int('3'+datetime.now().strftime('%Y%m%d')[2:])*100000
        simuaccount.crossday()
       
        
        
    def update(self,d):
        global g_dataSlide
        try:
            g_dataSlide[d['instid']] = d['data']
            self.curtime = datetime.strptime(str(d['data']['timedigit'])[:14],"%Y%m%d%H%M%S")
            #if not self.mode_724 and not is_valid_trade_time(d['instid'],curtime):
                #print("Invalid data time",curtime, d['instid'])
            #    return
            if d['data']['tradingday'] != self.curday:
                if self.curday != '':
                    self.crossday()
                self.curday = d['data']['tradingday']
            #print("update to matchTrade")
            self.matchTrade(d['instid'])
            
            simuaccount.current_timedigit = d['data']['timedigit']
            if simuaccount.tradingDay == '':
                self.loadFromDB(d['data']['tradingday'])
            
            simuaccount.tradingDay = d['data']['tradingday']
            
            if self.lasttime == 0 :
                #print('first update')
                self.lasttime = d['data']['timedigit']
                simuaccount.updateCapital(0, 0)
                simuaccount.saveToDB()
                
            elif abs(d['data']['timedigit'] - self.lasttime) > 2500:
                
                ##every above 2.5 second, update capital
                self.lasttime = d['data']['timedigit']
                simuaccount.updateCapital(0, 0)
                simuaccount.saveToDB()
            
        except Exception as e:
            logger.error(f"update error {e}",exc_info=True ) 
    
    def sendOrder(self,order):
        global g_dataSlide
        try:
            #if str(g_dataSlide[order['instid']]['timedigit'])[8:12] == '1015':
            #    order['status'] =  qetype.KEY_STATUS_REJECT
            #    order['cancelvol'] = order['leftvol']
            #    order['leftvol'] = 0
            #    order['errorid'] = 1
            #    order['errormsg'] =  'Not trade time.'
            #    #logger.warning(f'Warning: not trade time: {g_dataSlide[order['instid']]['timedigit']} ')                                   
            destprice = order['price'] if order['ordertype'] == 'limit' else g_dataSlide[order['instid']]['current']
            res = simuaccount.makeOrder(order['action'],order['direction'], destprice , order['volume'], order['instid'])
            
            if res != 0 :
                order['status'] =  'failed'
                order['cancelvol'] = order['leftvol']
                order['leftvol'] = 0
                order['errorid'] = abs(res)
                if res == -1:
                    order['errormsg'] =  '资金不足' 
                    logger.warning(f'Warning: 资金不足。剩余可用权益: {simuaccount.avail} ') 
                else:
                    order['errormsg'] =  "可平持仓不足"
                    instid = order['instid']
                    longpos = simuaccount.getPosition(instid,'long')
                    shortpos = simuaccount.getPosition(instid,'short')
                    logger.warning(f'Warning: 可平持仓不足。目前多仓 {longpos}, 空仓 {shortpos}') 
                    
#             else:
#                 logger.debug("pass account makeorder")

            #simuaccount.g_order_id += 1
            #self.mapTable[simuaccount.g_order_id] = order['incoming_orderid']
            #self.mapTable[order['incoming_orderid']] = simuaccount.g_order_id
            #order['orderid'] = simuaccount.g_order_id
            
            if order['status'] == 'committed':
                order['status'] = qetype.KEY_STATUS_PENDING
            elif order['status'] == 'failed':
                order['status'] =qetype.KEY_STATUS_REJECT
            else:
                order['status'] = qetype.KEY_STATUS_UNKNOWN
                
            simuaccount.orders[order['orderid']] = order
            #print(simuaccount.g_order_id, len(simuaccount.orders))
            self.rtnOrder(order)

            #self.matchTrade(instid)
        except Exception as e:
            logger.error(f"sendorder error {e}",exc_info=True ) 

            
    def cancelOrder(self,d):
        def docancel(order):
            if order['tradevol'] > 0:
                order['status'] = qetype.KEY_STATUS_PTPC
            else:    
                order['status'] = qetype.KEY_STATUS_CANCEL
            order['cancelvol'] = order['leftvol']
            order['leftvol'] = 0
            
        try:
            if d['orderid'] == 0:
                for id in simuaccount.orders.keys():
                    order = simuaccount.orders[id]
                    if order['stratName'] == d['stratName']:
                        if order['status'] == qetype.KEY_STATUS_PENDING or order['status'] == qetype.KEY_STATUS_PART_TRADED:
                            docancel(order)
                            simuaccount.cancelOrder( order['volume'], order['direction'], order['instid'],order['action'],saveDB=False)
                            self.rtnOrder(order)
                simuaccount.saveToDB()        
            else:
                order = simuaccount.orders.get(d['orderid'],None)
                if order:
                        if order['status'] == qetype.KEY_STATUS_PENDING or order['status'] == qetype.KEY_STATUS_PART_TRADED:
    #                         print('cancel order')
                            docancel(order)
                            self.rtnOrder(order)
                            simuaccount.cancelOrder( order['volume'], order['direction'], order['instid'],order['action'])
                        else:
                            logger.error('cancel order failed')

                else:
                        logger.error('cancel order is not found')

 
        except Exception as e:
            logger.error(f"cancelorder error {e}",exc_info=True ) 

    def rtnOrder(self,order):
        #global g_orders,g_order_id,mapTable
        d = {}
        d['type'] = qetype.KEY_ON_ORDER
        d['status'] = order['status']
        #print('rtnOrder', order['status'])
        content = simuaccount.orders.get(order['orderid'],-1)
        if content != -1:
            #stratID = int(str(incoming_orderid)[:5])
            d['stratName'] = order['stratName']
            #d['orderid'] = orderid
            d['tradevol'] = order['tradevol']
            d['leftvol'] = order['leftvol']
            d['cancelvol'] = order['cancelvol']
            
            temp_time = order.get('timedigit',-1)
            if temp_time != -1:
                d['timedigit'] = temp_time
            else:
                d['timedigit'] = int(datetime.now().strftime("%Y%m%d%H%M%S")+'001')
            
            time = str(d['timedigit'])
            d['time'] = time[:8]+' '+time[8:10]+':'+time[10:12]+':'+time[12:14]+"."+time[14:]
            order['time'] = d['time']
            d['orderid'] = order['orderid']
            
            simuaccount.orders[d['orderid']] = order
            sorder = order.copy()
            del sorder['autocancel']
            del sorder['autoremake']
            sorder['accid'] = 0
            saveOrderDataToDB(self.user,self.token, simuaccount.tradingDay, sorder )
            
            self.callback(d)
        else:
            logger.error('rsp order id is not found')

    def rtnTrade(self,trade):
        #global g_orders
        trade['type'] = qetype.KEY_ON_TRADE 
        inst = trade['instid']
        trade['presett'] = g_dataSlide[inst]['presett']
        self.callback(trade)
            
             
        
#             print('rsp trade id is not found')

    # def getStratName(self,data):
    #     tempdata = data.split("#")   # string is '',strateName,orderid
    #     return (tempdata[-2], tempdata[-1]) 
        

    def matchTrade(self,instid):
        global g_dataSlide
        
        tradeprice = 0.0
        tradevol = 0
        #debug_flag = False
        
        
        #print("matchTrade", self.g_order_id,self.g_benchmark)
        #return
        #print('matchTrade', len(simuaccount.orders))
        if len(simuaccount.orders) > 0:
            try:
                for key,order in simuaccount.orders.items():
                    #print(key, order)
                    if order['leftvol'] > 0 and instid == order['instid']:
#                         print('orderid '+str(order['incoming_orderid']))
                        #instid = order['instid']
                        #if order['status'] == qetype.KEY_STATUS_PART_TRADED:
                        #    print("check again:",order['status'], instid)
                        if not self.mode_724 and not is_valid_trade_time(instid,self.curtime):
                            print("MatchTrade failed on invalid data time",self.curtime, instid)
                            logger.info(f"MatchTrade failed on invalid data time:{self.curtime}, intid:{instid}")
                            continue
                        if g_dataSlide.get(instid, None) is None:
                            logger.info(f"MatchTrade failed on g_dataslide have no data on such {instid}" )
                            continue
                        if g_dataSlide[instid].get('presett',0) == 0:
                            logger.info(f"MatchTrade failed on presettle price is zero on such {instid}" )
                            continue
                            
                        stratName = order['stratName']
						 
                        found = False
                        if self.strats: 
                            if self.strats['async']:
	                            for strat in self.strats['strat']:
	                                if strat.name == stratName:
	                                    found = True                                   
	                                    flippage = strat.flippage
	                                    traderate = strat.traderate
	                                    feesmult = strat.feesmult
	                                    break
	                            if not found:
		                            flippage = 0
		                            traderate = 1
		                            feesmult = 1
                            else:
	                            if stratName in self.strats:
		                            strat = self.strats[stratName]['strat']
		                            flippage = strat.flippage
		                            traderate = strat.traderate
		                            feesmult = strat.feesmult
	                            else:
		                            flippage = 0
		                            traderate = 1
		                            feesmult = 1                            
                        ticksize = instSetts[instid]['ticksize']

                        tick = g_dataSlide[instid]
                        #print('matchTrade volume', instid, self.instVolume,tick['volume'])
                        opvol = tick['a1_v'] if order['direction'] > 0 else tick['b1_v']
                        
                        if instid in self.instVolume and tick['volume'] >= self.instVolume[instid]:
                                curvol = tick['volume'] - self.instVolume[instid]
                                self.instVolume[instid] = tick['volume']
                        else:
                            self.instVolume[instid] = tick['volume']
                            curvol = 0
                        
                        curvol = max(curvol, opvol)    
#                         logger.debug('curvol location '+str(curvol)+';')
                        
#                         if debug_flag:
#                             logger.debug('tick a='+str(tick['a1_p'])+', b='+str(tick['b1_p']))
#                             logger.debug('order left '+str(order['leftvol'])+' type '+str(order['ordertype'])+' type '+str(order['direction']))
                                
                        if order['ordertype'] == 'market':
#                             if debug_flag:
#                                 logger.debug('going to market')
                            if order['direction'] > 0:
                                
                                #tradeprice = tick['current'] if tick['current'] >= tick['a1_p'] else tick['a1_p']
                                #trade_able_count = tick['a1_v']
                                # waiting for next phase to address trade condition
                                tradeprice = tick['a1_p'] + flippage *ticksize if flippage > 0 and tradeprice < tick['a1_p'] + flippage*ticksize else tick['a1_p']
                            else:
                                #tradeprice = tick['current'] if tick['current'] <= tick['b1_p'] else tick['b1_p']
                                #trade_able_count = tick['b1_v']
                                # waiting for next phase to address trade condition
                                tradeprice = tick['b1_p'] - flippage *ticksize if flippage > 0 and tradeprice > tick['b1_p'] - flippage *ticksize else tick['b1_p']

                            #tradevol = max(int(curvol * traderate), 1)       
                            tradevol = order['leftvol']
                            #tradevol = order['leftvol']  if order['leftvol']  < tradevol else tradevol
#                             logger.debug('tradevol '+str(tradevol)+';')
                        elif order['ordertype'] == 'limit':
#                             if debug_flag:
#                                 logger.debug('going to limit')
                            tradeprice = order['price']
                            if (order['direction'] > 0 and tick['current'] > order['price'] ) or(order['direction'] < 0 and tick['current'] < order['price'] ):
                                tradevol = 0
                                #continue
                            elif tick['current'] == order['price']:
                                tradevol = curvol - order['pendvol'] if curvol >= order['pendvol'] else 0
                                order['pendvol'] -= order['pendvol'] if curvol > order['pendvol'] else curvol
                                if tradevol > 0:
                                    tradevol = order['leftvol']  if order['leftvol']  < tradevol else tradevol
                            else:
                                tradevol = max(int(curvol * traderate),1)
                                tradevol = order['leftvol']  if order['leftvol']  < tradevol else tradevol
                                
                            if (order['timecond'] =='FOK' and tradevol < order['leftvol']) or (order['timecond'] =='FAK' and tradevol == 0):
                                order['status'] = qetype.KEY_STATUS_CANCEL
                                order['tradevol'] = 0
                                order['cancelvol'] = order['volume']
                                order['leftvol'] = 0
                                order['timedigit'] = tick['timedigit']
                                self.rtnOrder(order)
                                tradevol = 0
                            '''   
                            elif (order['direction'] > 0 and tick['a1_p'] == order['price'] ) or(order['direction'] < 0 and tick['b1_p'] == order['price'] ):
                                tempvolume = tick['a1_v'] if order['direction'] > 0 else  tick['b1_v']
                                tradevol = order['leftvol'] if order['leftvol'] <= tempvolume else tempvolume
                            elif (order['direction'] > 0 and tick['a1_p'] < order['price'] ) or(order['direction'] < 0 and tick['b1_p'] > order['price'] ):
                                tradevol = order['leftvol'] # waiting for next phase to address liquidation
                            '''    

                        if tradevol > 0:
#                             print('matchTrade tradevol')
                            ##add closeYesterday codes
                            #if order['action'] == 'auto':
                            #    order['action'], tradevol = simuaccount.autoAction(instid, order['direction']>0, tradevol)
#                             logger.debug('autoaction tradevol '+str(tradevol))
                            closetype = order['closetype']
                            if order['action'] == 'close':
                                tradevol, closetype = simuaccount.checkCloseToday(instid, order['direction']<0, closetype, tradevol)
#                                 logger.debug('checkclosetoday tradevol '+str(tradevol))
                                if tradevol == 0:
                                    if order['timecond'] == 'FAK' or order['timecond'] == 'FOK':
                                        order['status'] = qetype.KEY_STATUS_CANCEL
                                        order['tradevol'] = 0
                                        order['cancelvol'] = order['volume']
                                        order['leftvol'] = 0
                                        order['timedigit'] = tick['timedigit']
                                        self.rtnOrder(order)
#                                     try:
#                                         print(simuaccount.position[instid]['long'])
#                                     except Exception as e:
#                                         print("print account position ",e.__traceback__.tb_lineno, e ) 
#                                     logger.debug('hit')
                                    #logger.warning(f'matchTrade invalid closetype {closetype}')
                                    continue
                            blong = order['direction']>0 if order['action']=='open' else order['direction']<0   
                            simuaccount.onTrade(instid, order['action'], blong, tradeprice, tradevol, closetype,feesmult)
                            
                            simuaccount.g_trade_id += 1
                            order['tradevol'] += int(tradevol)
                            order['leftvol'] = max(order['leftvol'] - int(tradevol),0)
                            order['timedigit'] = tick['timedigit']

                            if order['timecond'] == 'FAK':
                                order['status'] = qetype.KEY_STATUS_PTPC if order['leftvol'] > 0 else qetype.KEY_STATUS_ALL_TRADED
                                order['cancelvol'] = order['leftvol']
                                order['leftvol'] = 0
                            else:
                                if order['leftvol'] == 0:
                                    order['status'] = qetype.KEY_STATUS_ALL_TRADED
                                else:
                                    order['status'] = qetype.KEY_STATUS_PART_TRADED
                                
                            trade = {}

                            trade['instid'] = instid
                            trade['orderid'] = order['orderid']
                            trade['stratName'] = order['stratName']
                            trade['tradeid'] = simuaccount.g_trade_id                         
                            trade['tradevol'] = int(tradevol)
                            trade['tradeprice'] = tradeprice
                            trade['timedigit'] = tick['timedigit']
                            trade['date'] = tick['time'][:8]
                            trade['time'] = tick['time'][9:]
                            trade['action'] = order['action']
                            trade['dir'] = order['direction'] 
                            trade['closetype'] = closetype
                            logger.info(f"OnTrade:instid {instid},orderid {trade['orderid']},price {trade['tradeprice']},\
                                        vol {trade['tradevol']}, action {trade['action']},dir {trade['dir']}, closetype {trade['closetype']}")
                            trade['accid'] = 0
                            saveTradeDataToDB(self.user, self.token, simuaccount.tradingDay, trade )

                            if trade['stratName'] == 'force_close':
                                print(f"{tick['time']} Force Closed price:{trade['tradeprice']}, vol:{trade['tradevol']}, intid:{trade['instid']}, direction:{trade['dir']}, orderid: {trade['orderid']}, closetype:{trade['closetype']}")
                                
#                             tradedata = {}
                                                  
#                             tradedata['timedigit'] = tick['timedigit']
#                             tradedata['time'] = tick['time']
#                             tradedata['action'] = order['action']
#                             tradedata['dir'] = order['direction']     
#                             tradedata['tradeprice'] = tradeprice
#                             tradedata['vol'] = int(tradevol)
#                             tradedata['closetype'] = closetype
#                             trade['data'] = tradedata

                            simuaccount.trades[simuaccount.g_trade_id] = trade

                            self.rtnOrder(order)
                            self.rtnTrade(trade)

            except Exception as e:
                logger.error(f"matchTrade {e}",exc_info=True ) 
                        
                
        
        return





        
        
def runQETraderProcess(user, token, strats,traderqueue, mode_724):
    logger.info('start trader process')
    qesimtrader = QEsimtrader()
    #simuaccount.balance=balance
    #print('balance',self.balance)
    qesimtrader.user = user
    qesimtrader.token = token
    qesimtrader.strats = strats
    qesimtrader.tqueue = traderqueue
    #qesimtrader.loadFromDB()
    qesimtrader.mode_724 = mode_724
    #qesimtrader.master_orders = master_orders
    qesimtrader.simTraderProcess()






if __name__ == '__main__':
    #from multiprocessing import Queue
    #runQETraderProcess('scott','888888',Queue(),balance)
    pass












# -*- coding: utf-8 -*-
"""
Created on Wed Feb  1 19:07:43 2023

@author: ScottStation
"""


import json
import os

config_path = os.path.abspath(os.path.dirname(__file__)) +'/sysconfig.json'

def read_sysconfig():
    json_data = None
    if os.path.exists(config_path):
        with open(config_path,'r',encoding='utf8') as fp:
            json_data = json.load(fp)
            
    return json_data

def setRedisConfig(host='127.0.0.1', port=6379, password=''):
    try:
        assert isinstance(host, str) and isinstance(password, str), 'host和password必须是合法ip地址'
        assert isinstance(port, int), 'port 必须是int类型'
        json_data = read_sysconfig()
        assert not json_data is None, 'json配置数据文件丢失'
        json_data['redis'] ={'host': host,'port':port,'password':password}
        with open(config_path,'w') as fp:
            json.dump(json_data, fp)
    except Exception as e:
        print(f'Error:{e.__traceback__.tb_lineno},{e}')
        
def setWebConfig(host='127.0.0.1',port=5814):
    try:
        assert isinstance(host, str), 'host必须是合法ip地址'
        assert isinstance(port, int), 'port 必须是int类型'
        json_data = read_sysconfig()
        assert not json_data is None, 'json配置数据文件丢失'
        json_data['webpage'] ={'host': host,'port':port}
        with open(config_path,'w') as fp:
            json.dump(json_data, fp)
    except Exception as e:
        print(f'Error:{e.__traceback__.tb_lineno},{e}')


if __name__ == "__main__":
    setWebConfig()
    print(read_sysconfig())
    



# -*- coding: utf-8 -*-
"""
Created on Fri May 20 22:38:14 2022

@author: ScottStation
"""
import numpy as np
import pandas as pd

import scipy.stats as si
#from scipy.stats import norm
#N = norm.cdf

def get_d(S, K, T, r, vol):
    d1 = (np.log(S / K) + (r + 0.5 * vol ** 2) * T) / (vol * np.sqrt(T))
    d2 = d1 - vol * np.sqrt(T)
    return (d1,d2)   


def bs_call(S, K, T, r, vol):
    (d1,d2) = get_d(S, K ,T, r, vol)
    return S * si.norm.cdf(d1) - np.exp(-r * T) * K * si.norm.cdf(d2)

def bs_put(S, K, T, r, vol):
    (d1,d2) = get_d(S, K ,T, r, vol)
    return np.exp(-r * T) * K * si.norm.cdf(-d2) - S * si.norm.cdf(-d1)

def bs_vega(S, K, T, r, vol):
	d1 = get_d(S,K,T,r,vol)[0]
	return S * si.norm.pdf(d1) * np.sqrt(T) 

def bs_delta(S, K, T, r, vol, isCall):
	d1 = get_d(S,K,T,r,vol)[0]
	n = 1 if isCall else -1
	return  n * si.norm.cdf(n * d1)
	
def bs_gamma(S, K, T, r, vol):
	d1 = get_d(S,K,T,r,vol)[0]
	return  si.norm.pdf(d1) / (S * vol * np.sqrt(T))
  

def bs_theta(S, K, T, r, vol, isCall):
	(d1,d2) = get_d(S,K,T,r,vol)
	n = 1 if isCall else -1
	return  (-1 * (S * si.norm.pdf(d1) * vol) / (2 * np.sqrt(T)) - n * r * K * np.exp(-r * T) * si.norm.cdf(n * d2)) / 365
    

def find_vol(target_value, S, K, T, r, isCall=True):
    MAX_ITERATIONS = 200
    PRECISION = 1.0e-5
    sigma = 0.3
    for i in range(0, MAX_ITERATIONS):
        price = bs_call(S, K, T, r, sigma) if isCall else bs_put(S,K,T,r,sigma)
        vega = bs_vega(S, K, T, r, sigma)
        diff = target_value - price  # our root
        if (abs(diff) < PRECISION):
            return sigma
        sigma = sigma + diff/vega # f(x) / f'(x)
    return sigma # value wasn't found, return best guess so far


# -*- coding: utf-8 -*-
"""
Created on Thu Mar 24 21:11:45 2022

@author: ScottStation
"""
from .qelogger import logger
#from .qeriskctl import soptriskctl, ctpriskctl
from .qetype import qetype
from .qeredisdb import saveHedgePointToDB, saveHedgePointrealToDB, get_bar_data
from multiprocessing import Queue
import pandas as pd
import numpy as np
#import datetime
from .qesimtrader import getMixPrice
from .qestatistics import g_stat
from .qeglobal import getAccidTraderQueue, getInstTraderQueue, getInstClass, getInstAccID, getClassAccID
from datetime import datetime, timedelta
import traceback
import random
import asyncio


simuqueue = Queue()

class qeStratBase:
    flippage = 0
    traderate = 1.0
    formula = None
    wait_all=False
    freq=0
    instid = None
    datamode ='tick'
    rfrate = 0.00
    
    feesmult = 1.0
    recordinsts = None
    direction_index = 0
    dailyminutes = 5
    
    _runmode = 'simu'
    _start_date = None
    _end_date = None
    _append_mode = False
    
    @classmethod
    def setMode(cls, runmode, start_date=None, end_date=None):
        cls._runmode = runmode
        cls._start_date = start_date
        cls._end_date = end_date

 
    def handleData(self, context):
        pass
    
    def onBar(self,context):
        pass

    def crossDay(self, context):
        pass
    
    def handleOrder(self,order, context):
        pass

    def handleOrderError(self,order, context):
        pass
    
    def handleTrade(self,trade,context):
        pass

    async def aio_onBar(self,context):
        pass
    
    async def aio_crossDay(self,context):
        pass

    def matchTrade(self, context, datamode):
        try:
            # trades = {'id':{'tradeprice}}
            tradeprice = 0.0
            tradevol = 0
            orderinstid = ''
            trades = {}
            flippage = self.flippage
            traderate = self.traderate
            # print(context.orders)
            for oid in context.orders:
                orderinstid = context.orders[oid]['instid']
                if context.orders[oid]['leftvol'] == 0:
                    continue
                ## dataslide trade process
                elif datamode == 'tick':
                    current = context.dataslide[orderinstid]['current']
                    if context.orders[oid]['ordertype'] == 'market':
                        if context.orders[oid]['direction'] > 0:
                            tradeprice = current if current >= context.dataslide[orderinstid]['a1_p'] else \
                                context.dataslide[orderinstid]['a1_p']
                            tradeprice = context.dataslide[orderinstid]['a1_p'] + flippage * \
                                         context.instsett[orderinstid][
                                             'ticksize'] if flippage > 0 and tradeprice < \
                                                            context.dataslide[orderinstid][
                                                                'a1_p'] + flippage * context.instsett[orderinstid][
                                                                'ticksize'] else tradeprice
                        else:
                            tradeprice = current if current <= context.dataslide[orderinstid]['b1_p'] else \
                                context.dataslide[orderinstid]['b1_p']
                            tradeprice = context.dataslide[orderinstid]['b1_p'] - flippage * \
                                         context.instsett[orderinstid][
                                             'ticksize'] if flippage > 0 and tradeprice > \
                                                            context.dataslide[orderinstid][
                                                                'b1_p'] - flippage * context.instsett[orderinstid][
                                                                'ticksize'] else tradeprice
                        #tradevol = max(int(context.curvol[orderinstid] * traderate), 1)
                        #tradevol = context.orders[oid]['leftvol'] if context.orders[oid][
                        #                                                'leftvol'] < tradevol else tradevol
                        tradevol =  context.orders[oid]['leftvol']

                    elif context.orders[oid]['ordertype'] == 'limit':
                        tradeprice = context.orders[oid]['price']
                        if (context.orders[oid]['direction'] > 0 and current > context.orders[oid]['price']) or (
                                context.orders[oid]['direction'] < 0 and current < context.orders[oid]['price']):
                            tradevol = 0
                            continue
                        elif current == context.orders[oid]['price']:
                            tradevol = context.curvol[orderinstid] - context.orders[oid]['pendvol'] if context.curvol[
                                                                                                          orderinstid] >= \
                                                                                                      context.orders[
                                                                                                          oid][
                                                                                                          'pendvol'] else 0
                            context.orders[oid]['pendvol'] -= context.orders[oid]['pendvol'] if context.curvol[
                                                                                                  orderinstid] > \
                                                                                              context.orders[oid][
                                                                                                  'pendvol'] else \
                                context.curvol[orderinstid]
                            if tradevol == 0:
                                continue
                            else:
                                tradevol = context.orders[oid]['leftvol'] if context.orders[oid][
                                                                                'leftvol'] < tradevol else tradevol
                        else:
                            tradevol = max(int(context.curvol[orderinstid] * traderate), 1)
                            tradevol = context.orders[oid]['leftvol'] if context.orders[oid][
                                                                            'leftvol'] < tradevol else tradevol

                ## minu data trade process

                elif datamode == 'minute':
                    
                    current = context.dataslide[orderinstid]['close']
                    if context.orders[oid]['ordertype'] == 'market':
                        if context.orders[oid]['direction'] > 0:
                            tradeprice = current + flippage * context.instsett[orderinstid]['ticksize']
                        else:
                            tradeprice = current - flippage * context.instsett[orderinstid]['ticksize']
#                         print(type(context.curvol[orderinstid]))
#                         print(type(traderate))
                        #tradevol = max(1,int(context.curvol[orderinstid] * traderate))
                        tradevol = context.orders[oid]['leftvol']

                    elif context.orders[oid]['ordertype'] == 'limit':
                        tradeprice = context.orders[oid]['price']
                        # spread = 2 * context.instsett[orderinstid]['ticksize']
                        if (context.orders[oid]['direction'] > 0 and context.dataslide[orderinstid]['close'] >
                            context.orders[oid]['price']) or (
                                context.orders[oid]['direction'] < 0 and context.dataslide[orderinstid]['close'] <
                                context.orders[oid]['price']):
                            tradevol = 0
                            continue
                        else:
                            tradevol = max(1, int(context.curvol[orderinstid] * traderate))
                    tradevol = context.orders[oid]['leftvol'] if context.orders[oid]['leftvol'] < tradevol else tradevol

                ##minute and daily trade
                else:
                    current = context.dataslide[orderinstid]['close']
                    if context.orders[oid]['ordertype'] == 'market':
                        if context.orders[oid]['direction'] > 0:
                            tradeprice = current + flippage * context.instsett[orderinstid]['ticksize']
                        else:
                            tradeprice = current - flippage * context.instsett[orderinstid]['ticksize']
                        #tradevol = max(1,int(context.curvol[orderinstid] * traderate))
                        tradevol = context.orders[oid]['leftvol']

                    elif context.orders[oid]['ordertype'] == 'limit':
                        tradeprice = context.orders[oid]['price']
                        if (context.orders[oid]['direction'] > 0 and context.dataslide[orderinstid]['close'] >
                            context.orders[oid]['price']) or (
                                context.orders[oid]['direction'] < 0 and context.dataslide[orderinstid]['close'] <
                                context.orders[oid]['price']):
                            tradevol = 0
                            continue
                        else:
                            tradevol = max(1,int(context.curvol[orderinstid] * traderate))
                    tradevol = context.orders[oid]['leftvol'] if context.orders[oid]['leftvol'] < tradevol else tradevol
                if (context.orders[oid]['timecond']== 'FAK' and tradevol == 0) or \
                   (context.orders[oid]['timecond']== 'FOK' and tradevol < context.orders[oid]['volume']):
                    tradevol = 0
                    context.orders[oid]['cancelvol'] = context.ordere[oid]['leftvol']
                    context.orders[oid]['leftvol'] = 0
                    context.orders[oid]['status'] = 'canceled'

                if tradevol > 0:
                    # print(context.curtime,orderinstid,oid)
                    # tradeinstid = orderinstid
                    trades[oid] = {'tradeprice': tradeprice, 'tradevol': int(tradevol), 'instid': orderinstid}
        except Exception as e:
            print("Failed on matchTrade: ", e.__traceback__.tb_lineno, e)

        return trades

def real_make_order(context, instid, direction, price, volume, ordertype="limit", action="open", closetype='auto',timecond='GFD', autoremake=[0,0,0,0], autocancel=[0,0],accid=-1,father=0):
    '''
    Call this function to make an order
    Paramters
    --------
    context: object
        Context data of this backtest
    direction: int
        Long 1, short -1
    price: float
        The price of this order (except market order)
    volume: int
        The volume of this order
    ordertype: string
        Only support following types: "market" for market orders and "limit" for limit orders.
    action: string
        Only support "open" for open only, "close" for close only.
    closetype: string
        default: 'auto ' for automatic choice
        also could be 'closetoday' or 'closeyesterday' 
    Returns
    -------
    -1 if failed, otherwisa return order oid (>0).

    '''
    #global traderqueue
    #traderqueue = getInstTraderQueue(instid) if context.runmode=='real' else simuqueue
    validtypes = ['limit', 'market']
    validtimeconds = ['GFD', 'FAK', 'FOK']
    validclosetypes = ['auto', 'closetoday', 'closeyesterday']
    validaction = ['open', 'close']  # ['auto','open','close']
    try:
        direction = int(direction)
        volume = int(volume)
    except Exception as e:
        logger.warning("volume/direction must be digits." + str(direction) + '/' + str(volume), e)
        return -1,[]
    if not ordertype in validtypes:
        logger.warning("ordettype must be one of " + str(validtypes) + ",your ordertype:" + str(ordertype))
        return -1,[]
    if not closetype in validclosetypes:
        logger.warning("closetype must be one of " + str(validclosetypes) + ',your closetype:' + str(closetype))
        return -1,[]
    if not timecond in validtimeconds:
        logger.warning("closetype must be one of " + str(validclosetypes) + ',your closetype:' + str(closetype))
        return -1,[]
    if not action in validaction:
        logger.warning("action must be one of " + str(validaction) + ',your action:' + str(action))
        return -1,[]

    if direction == 0:
        logger.warning("Direction can not be zero. your direction: " + str(direction))
        return -1,[]
    if ordertype =='market' and price == 0:
        price = context.getCurrent(instid)
    if volume <= 0 or price <= 0:
        logger.warning("price and volume must be positive . current price:" + str(price) + ',volume:' + str(volume))
        return -1,[]
    if autoremake[0] == 1 and timecond != 'FAK':
        logger.warning(f"FAK track mode 1 must set timecond to FAK but current is {timecond}")
        return -1,[]
    if autoremake[0] == 2 and timecond != 'GFD':
        logger.warning(f"FAK track mode 2 must set timecond to GFD but current is {timecond}")
        return -1,[]

    #if action == 'close' and closetype == 'none':
    #    logger.warning('closetype at close must be close,closetoday or closeyesterday')
    #    return -1

    #if action == 'close' and closetype == 'close':

    #    if instid[-3:] == 'SFE' or instid[-3:] == 'INE':
    #        logger.warning('closetype must be closetoday or closeyesterday for instrument in SFE or INE')
    #        return -1

    classname = getInstClass(instid)
    accid = getInstAccID(instid)[0] if accid < 0 else accid
    if accid >= len(context.accounts):
        accid = len(context.accounts)-1
    traderqueue = getAccidTraderQueue(accid) if context.runmode=='real' else simuqueue    
    #    if soptriskctl.newCount() == -1:
    #        logger.warning("Risk Control: order numbers exceed maxium of a day.")
    #        return -2
    #    elif soptriskctl.newCount() == -2:
    #        logger.warning('Risk Control: order numbers exceed maxium of a second, try later.')
    #        return -3 
        
        
    try:

        status = 'committed'
        errorid = 0
        errormsg = ''
        #leftvol = volume
        cancelvol = 0
        #destprice = price if ordertype == 'limit' else context.getCurrent(instid)
        autoremake += [datetime.now(), 'wait'] 
        autocancel += [datetime.now(), 'wait']
        
        if True: # not context.realTrade:
            ''' #move to simtrader
            margin = context.getMargin(direction>0,volume,instid)
            #print(margin, context.stat.avail, context.stat.balance)
            if context.stat.avail <= margin:
                    status = 'failed'
                    errorid  = 1
                    errormsg = 'Not enough margin.'
                    leftvol = 0
                    cancelvol = volume
                    print("Warning: not enough capital for new order. Current availiable:", context.stat.avail, 'margin:',margin)
            
            if status == 'committed' and action == 'open':
                if direction > 0:
                    context.longpendvol[instid]['poscost'] = getMixPrice(context.longpendvol[instid]['poscost'],context.longpendvol[instid]['volume'],destprice, volume)
                    context.longpendvol[instid]['volume'] +=  volume
                else:
                    context.shortpendvol[instid]['poscost'] = getMixPrice(context.shortpendvol[instid]['poscost'],context.shortpendvol[instid]['volume'],destprice, volume)                            
                    context.shortpendvol[instid]['volume'] += volume
                context.updateFrozenMarg()
'''
        pendvol = 0
        if ordertype == 'limit':
            pendvol = context.getDataSlide(instid,'a1_v') if direction < 0 else context.getDataSlide(instid,'b1_v')
        
        if instid[-3:] == 'CCF':
                ## CCFX order commission
                context.stat.dayfees += 1
                context.stat.totalfees += 1
        context.stat.avail = context.stat.balance - context.marg - context.frozenMarg
        context.stat.dayorders += 1
        
        orderids = []
        if action == 'close' and closetype=='auto':
            dirstr = 'long' if  direction < 0 else 'short'
            yesvol = context.getAccountPosition(instid, dirstr, 'yesvol')
            tvol = context.getAccountPosition(instid, dirstr, 'volume')
            if yesvol > 0:
                if volume > yesvol:
                    closevol = [ min(volume - yesvol, tvol - yesvol), yesvol]
                else:
                    closevol = [ 0, volume]
            else:
                closevol = [min(volume, tvol - yesvol), 0]
            
            ordernum = 2 if closevol[0] != 0 and closevol[1] != 0 else 1
            if ordertype == 'limit':
                rc = context.accounts[accid].riskctl.make_order(context, ordernum, instid, price, direction)
            else:    
                rc = context.accounts[accid].riskctl.make_order(context, ordernum)
            #rc = context.accounts[accid].riskctl.make_order(context, ordernum)
            if rc < 0:
                logger.warning(f"风控模块阻止下单: error code {rc} on {instid} {action} {direction}")
                return rc,[]

            for i in range(len(closevol)):
                if closevol[i] > 0:
                    context.orderid = g_stat.getNewOrderID()
                    closetype = 'closetoday' if i ==0 else 'closeyesterday'
                    temporder = {'instid': instid,
                                 'price': price,
                                 'direction': direction,
                                 'ordertype': ordertype,
                                 'closetype': closetype,
                                 'volume': closevol[i],
                                 'leftvol': closevol[i],
                                 'tradevol': 0,
                                 'cancelvol': cancelvol,
                                 'pendvol': pendvol,
                                 'status': status,
                                 'errorid': errorid,
                                 'errormsg': errormsg,
                                 'action': action,
                                 'timecond': timecond,
                                 'autoremake':autoremake,
                                 'autocancel': autocancel,
                                 'father': father,
                                 'accid':accid}
            
                    context.orders[context.orderid] = temporder
                    temporder['type'] = qetype.KEY_SEND_ORDER
                    temporder['stratName'] = context.stratName
                    temporder['orderid'] = context.orderid
                    orderids.append(context.orderid)
                    traderqueue.put(temporder)
        else:            
            if ordertype == 'limit':
                rc = context.accounts[accid].riskctl.make_order(context, 1, instid, price, direction)
            else:    
                rc = context.accounts[accid].riskctl.make_order(context, 1)
            #rc = context.account.riskctl.make_order(context)
            if rc < 0:
                logger.warning(f"风控模块阻止下单: error code {rc} on {instid} {action} {direction}")
                return rc,[]
            context.orderid = g_stat.getNewOrderID()
            temporder = {'instid': instid,
                         'price': price,
                         'direction': direction,
                         'ordertype': ordertype,
                         'closetype': closetype,
                         'volume': volume,
                         'leftvol': volume,
                         'tradevol': 0,
                         'cancelvol': cancelvol,
                         'pendvol': pendvol,
                         'status': status,
                         'errorid': errorid,
                         'errormsg': errormsg,
                         'action': action,
                         'timecond': timecond,
                         'autoremake':autoremake,
                         'autocancel': autocancel,
                         'father': father,
                         'accid':accid}
    
            context.orders[context.orderid] = temporder
            temporder['type'] = qetype.KEY_SEND_ORDER
            temporder['stratName'] = context.stratName
            temporder['orderid'] = context.orderid
            orderids.append(context.orderid)
            traderqueue.put(temporder)
    
        return 0, orderids
    except Exception as e:
        logger.error("Failed on make order: " + str(e), exc_info=True)
        return -1,[]


def real_cancel_order(context, orderid):
    '''
    Paramters
    --------
    context: object
        Context data of this backtest
    orderid: int
        The order oid. if 0 means cancel all orders.
    Returns
    -------
    -1 failed to find orderid
    0 successed

    '''
    #global traderqueue
    traderqueue = None 
   
    try:
        ret = -1
        if orderid == 0:
            totals = {}
            totalnum = 0
            for oid in context.orders:
                if context.orders[oid]['leftvol'] > 0:
                    accid = context.orders[oid]['accid']
                    totalnum += 1
                    if not accid in totals:
                        totals[accid] = 1
                    else:
                        totals[accid] += 1
                        
            for accid in totals:
                rc = context.accounts[accid].riskctl.cancel_order(context, totals[accid])
                if rc  < 0:
                    logger.warning(f'风控模块阻止撤单: error code:{rc}')
                    return rc
            context.frozenMarg = 0
            #for instid in context.instid:
            #    context.longpendvol[instid] = {'volume':0,'poscost':0}
            #    context.shortpendvol[instid] = {'volume':0,'poscost':0}
            for oid, order in context.orders:
                instid = order['instid']
                ## set the remake count to zero
                context.orders[oid]['autoremake'][3] = 0 

                if instid[-3:] == 'CCF':
                    ## CCFX cancel commission
                    context.stat.dayfees += 1
                    context.stat.totalfees += 1
            ret = 0
            context.stat.daycancels += totalnum

            #     d = {}
            #     d['type'] = KEY_CANCEL_ORDER
            #     d['stratName'] = context.stratName
            #     d['orderid'] = order['orderid']
            #     traderqueue.put(d)
            print(context.curtime, 'cancel all order')
            logger.info('cancel all order')
            context.stat.avail = context.stat.balance - context.marg - context.frozenMarg
                # context.Cancels[context.curtime] = [orderid, context.stat.daycancels]
                # print(context.Cancels[context.curtime])
            temporder = {}
            temporder['type'] = qetype.KEY_CANCEL_ORDER
            temporder['stratName'] = context.stratName
            temporder['orderid'] = orderid
            if context.runmode =='real':
                for accid in totals:
                    if totals[accid] > 0:
                        getAccidTraderQueue(accid).put(temporder)
            else:
                simuqueue.put(temporder)
            # return ret

        else:
            

            if orderid in context.orders.keys():
                instid = context.orders[orderid]['instid']
                accid = context.orders[orderid]['accid']
                rc = context.accounts[accid].riskctl.cancel_order(context, 1)
                if rc  < 0:
                    logger.warning(f'风控模块阻止撤单 {orderid}: error code:{rc}')
                    return rc
                #if True : #not context.realTrade:
                traderqueue = getAccidTraderQueue(accid)
                #if classname == 'stockoption':
                #    if soptriskctl.newCount() == -1:
                #        logger.warning("Risk Control: order numbers exceed maxium of a day.")
                #        return -2
                #    elif soptriskctl.newCount(1, [orderid]) == -2:
                #        logger.warning('Risk Control: order numbers exceed maxium of a second, try later.')
                #        return -3 
                #elif classname == 'future':
                #    if ctpriskctl.cancelCount() < 0:
                #        logger.warning('Risk Control: cancel numbers exceeed maxium of a day.')
                #        return -2
                
                '''
                if context.orders[orderid]['action'] == 'open':
                    if context.orders[orderid]['direction'] > 0:
                        context.longpendvol[instid]['volume'] = max(
                            context.longpendvol[instid]['volume'] - context.orders[orderid]['leftvol'], 0)
                        if context.longpendvol[instid]['volume'] <= 0:
                            context.longpendvol[instid] = {'volume':0,'poscost':0}
                    else:
                        context.shortpendvol[instid]['volume'] = max(
                            context.shortpendvol[instid]['volume'] - context.orders[orderid]['leftvol'], 0)
                        if context.shortpendvol[instid]['volume'] <= 0:
                            context.shortpendvol[instid] = {'volume':0,'poscost':0}
                    context.updateFrozenMarg()
                '''    
                if not context.realTrade:    
                    context.orders[orderid]['status'] = "canceled" if context.orders[orderid][
                                                                          'status'] != "parttraded" else "ptpc"
                    # part traded part canceled
                    context.orders[orderid]['cancelvol'] = context.orders[orderid]['leftvol']
                    context.orders[orderid]['leftvol'] = 0
                if instid[-3:] == 'CCF':
                    ## CCFX order commission
                    context.stat.dayfees += 1
                    context.stat.totalfees += 1
                context.stat.daycancels += 1
                ret = 0
            if ret == 0 and traderqueue:
                context.stat.avail = context.stat.balance - context.marg - context.frozenMarg
                # context.Cancels[context.curtime] = [orderid, context.stat.daycancels]
                # print(context.Cancels[context.curtime])
                
                ## set the remake count to zero
                #context.orders[orderid]['autoremake'][3] = 0 
                logger.info(f'{ context.orders[orderid]["instid"]} Cancel order:{orderid}')
                print(f'{context.curtime} { context.orders[orderid]["instid"]} Cancel order:{orderid}')
                temporder = {}
                temporder['type'] = qetype.KEY_CANCEL_ORDER
                temporder['stratName'] = context.stratName
                temporder['orderid'] = orderid
                if context.runmode =='real':
                    traderqueue.put(temporder)
                else:
                    simuqueue.put(temporder)
        return ret
    except Exception as e:
        logger.error("Failed on cancel order: ", e.__traceback__.tb_lineno, e)
        return -1


def calculatemul(context):
    letters = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u',\
                 'v', 'w', 'x', 'y', 'z']
    mapDict = {}
    for i in range(len(context.instid)):
        mapDict[letters[i]] = context.getCurrent(context.instid[i])
    # print(context.formula)
    multiresult = eval(context.formula, mapDict)
    # print(multiresult)
    return multiresult

def real_record_hedge_point(context, action, dirstr, volume):
    validaction = ['open', 'close']  # ['auto','open','close']
    valid_direction = ['long', 'short']

    if not action in validaction:
        logger.error("action must be one of " + str(validaction) + ',your action:' + str(action))
        return

    if not dirstr in valid_direction:
        logger.error("action must be one of " + str(valid_direction) + ',your direction:' + str(valid_direction))
        return

    if context.hedgemodel == False:
        return

    d = {}
    d['action'] = action
    d['direction'] = dirstr
    d['volume'] = volume
    d['time'] = context.timedigit
    d['stratName'] = context.stratName
    d['price'] = calculatemul(context)
    #if context.realTrade:
    #    saveHedgePointrealToDB(context.user, context.token, context.stratName, context.timedigit, d)
    #    logger.info('write hedge point to DB')

    #else:
    saveHedgePointToDB(context.user, context.token, context.stratName, context.timedigit, d)
    logger.info('write hedge point to DB')
    return




def test_make_order(context, instid, direction, price, volume, ordertype="limit", action="open", closetype='auto', timecond = 'GFD'):
    '''
    Call this function to make an order
    Paramters
    --------
    context: object
        Context data of this backtest
    direction: int
        Long 1, short -1
    price: float
        The price of this order (except market order)
    volume: int
        The volume of this order
    ordertype: string
        Only support following types: "market" for market orders and "limit" for limit orders.
    action: string
        Only support "open" for open only, "close" for close only.
    closetype: string
        default: 'auto' for automatic match
        also could be 'closetoday' or 'closeyesterday'
    Returns
    -------
    -1 if failed, otherwisa return order oid (>0).

    '''
    # print(1)
    validtypes = ['limit', 'market']
    validclosetypes = ['auto', 'closetoday', 'closeyesterday'] 
    validtimeconds = ['GFD', 'FAK', 'FOK']
    validaction = ['open', 'close']
    

    try:
        direction = int(direction)
        volume = int(volume)
        # print(direction,volume)
        # print(instid)
    except Exception as e:
        logger.error(f"volume/direction must be digits.,{direction},{volume}, {e}")
        return -1,[]
    if not ordertype in validtypes:
        logger.warning(f"ordettype must be one of {validtypes}, your ordertype:{ordertype}")
        return -1,[]
    if not action in validaction:
        logger.warning(f"action must be one of {validaction},your action:{ action}")
        return -1,[]
    if not closetype in validclosetypes:
        logger.warning(f"action must be one of {validclosetypes},your action:{ closetype}")
        return -1,[]
    if not timecond in validtimeconds:
        logger.warning("closetype must be one of " + str(validclosetypes) + ',your closetype:' + str(closetype))
        return -1,[]
    if direction == 0:
        logger.warning(f"Direction can not be zero. your direction: {direction}")
        return -1,[]
    if ordertype =='market' and price == 0:
        price = context.getCurrent(instid)
    if volume <= 0 or price <= 0:
        logger.warning(f"price and volume must be positive . current price:{price}, volume:{volume},instid:{instid}")
        return -1,[]
    
    #if action == 'close' and closetype=='none':
    #        logger.warning('closetype must be close,closetoday or closeyesterday if action=close')
    #        return -1
        
    #if action == 'close' and closetype=='close':
        
    #    if instid[-3:] == 'SFE' or instid[-3:] == 'INE':
    #        logger.warning('closetype must be closetoday or closeyesterday for instrument in SFE or INE')      
    #        return -1
        

    try:
        status = 'committed'
        errorid = 0
        errormsg = ''
        leftvol = volume
        cancelvol = 0
        destprice = price if ordertype=="limit" else context.getCurrent(instid)
        if action == "open":
            
            margin = context.getMargin(destprice, direction > 0, volume, instid)
            # print(margin, context.stat.avail, context.stat.balance)
            if context.stat.avail <= margin:
                status = 'failed'
                errorid = 1
                errormsg = '资金不足.'
                leftvol = 0
                cancelvol = volume
                logger.warning(f"Warning: 资金不足，目前可用权益:{context.stat.avail}, 需要保证金:{margin}")
            
        elif action == "close":
            dirstr = 'short' if  direction > 0 else 'long'
            pend = context.calcFrozenVol(instid,direction)
            pos = context.position[instid][dirstr]['volume']
            leftvol = pos - pend
            if volume > leftvol:
                status = 'failed'
                errorid = 1
                errormsg = '可平持仓不足.'
                leftvol = 0
                cancelvol = volume
                logger.warning(f"Warning: 可平持仓不足. 目前持仓:{pos}, 冻结持仓:{pend}")
                #elif instid in context.frozenVol:
                #    context.frozenVol[instid]['short'] += volume;
                #else:
                #    context.frozenVol[instid] = {"long":0, "short":volume}

                   
                    
        if errorid == 0:
            if status == 'committed' and action == 'open':
                '''
                if direction > 0:
                    context.longpendvol[instid]['poscost'] = getMixPrice(context.longpendvol[instid]['poscost'],context.longpendvol[instid]['volume'],destprice, volume)
                    context.longpendvol[instid]['volume'] +=  volume
                else:
                    
                    context.shortpendvol[instid]['poscost'] = getMixPrice(context.shortpendvol[instid]['poscost'],context.shortpendvol[instid]['volume'],destprice, volume)                            
                    context.shortpendvol[instid]['volume'] += volume
                '''    
            context.updateFrozenMarg()
            context.stat.avail = context.stat.balance - context.marg - context.frozenMarg
    
        if context.datamode == 'tick':
            pendvol = context.dataslide[instid]['a1_v'] if direction < 0 else context.dataslide[instid]['b1_v']
        else:
            pendvol = 0
        if errorid == 0 and context.instid[-3:] == 'CCF':
            ## CCFX order commission
            context.stat.dayfees += 1
            context.stat.totalfees += 1
            
        context.stat.dayorders += 1
        
        
        orderids = []
        
        if action =='close' and closetype=='auto' and errorid == 0:
            dirstr = 'long' if  direction < 0 else 'short'
            yesvol = context.getAccountPosition(instid, dirstr, 'yesvol')
            tvol = context.getAccountPosition(instid, dirstr, 'volume')
            if yesvol > 0:
                if volume > yesvol:
                    closevol = [ min(volume - yesvol, tvol - yesvol), yesvol]
                else:
                    closevol = [ 0, volume]
            else:
                closevol = [min(volume, tvol - yesvol), 0]
        
            for i in range(len(closevol)):
                if closevol[i] > 0:
                    context.orderid += 1
                    # pendvol = 0
                    # if ordertype == 'limit':
                    closetype = 'closetoday' if i == 0 else 'closeyesterday'    
                    context.orders[context.orderid] = {'instid': instid,
                                                       'time': context.curtime,
                                                       'price': price,
                                                       'direction': direction,
                                                       'ordertype': ordertype,
                                                       'closetype': closetype,
                                                       'volume': closevol[i],
                                                       'leftvol': leftvol,
                                                       'tradevol': 0,
                                                       'cancelvol': cancelvol,
                                                       'pendvol': pendvol,
                                                       'status': status,
                                                       'errorid': errorid,
                                                       'errormsg': errormsg,
                                                       'timecond': timecond,
                                                       'action': action, }
                    orderids.append(context.orderid)
        else:
            context.orderid += 1
            # pendvol = 0
            # if ordertype == 'limit':
            context.orders[context.orderid] = {'instid': instid,
                                               'time': context.curtime,
                                               'price': price,
                                               'direction': direction,
                                               'ordertype': ordertype,
                                               'closetype': closetype,
                                               'volume': volume,
                                               'leftvol': leftvol,
                                               'tradevol': 0,
                                               'cancelvol': cancelvol,
                                               'pendvol': pendvol,
                                               'status': status,
                                               'errorid': errorid,
                                               'errormsg': errormsg,
                                               'timecond': timecond,
                                               'action': action, }
            orderids.append(context.orderid)
        #context.ROrders[context.curtime] = [context.orderid, price, direction, volume, ordertype, closetype, action,
        #                                    errorid, errormsg]
        # print(context.ROrders[context.curtime])
        #print('make_order',volume, action,closetype,leftvol)
        return 0,orderids
    except Exception as e:
        traceback.print_exc()
        logger.error(f"Failed on make order: {e.__traceback__.tb_lineno}, {e}")
        return -1,[]


def test_cancel_order(context, orderid):
    '''
    Paramters
    --------
    context: object
        Context data of this backtest
    orderid: int
        The order oid. if 0 means cancel all orders.
    Returns
    -------
    -1 failed to find orderid
    0 successed

    '''
    try:
        ret = -1
        if orderid == 0:
            ordernum = len(context.orders)
            #context.orders[orderid] = {}
            context.frozenMarg = 0
            #for instid in context.instid:
            #    context.longpendvol[instid] = {'volume':0,'poscost':0}
            #    context.shortpendvol[instid] = {'volume':0,'poscost':0}
            for orderid in context.orders.keys():
                inst = context.orders[orderid]['instid']
                if inst[-3:] == 'CCF':
                    ## CCFX cancel commission
                    context.stat.dayfees += 1
                    context.stat.totalfees += 1
                context.orders[orderid]['status'] = "canceled" if context.orders[orderid][
                                                                      'status'] != "parttraded" else "ptpc"
                # part traded part canceled
                context.orders[orderid]['cancelvol'] = context.orders[orderid]['leftvol']
                context.orders[orderid]['leftvol'] = 0
            ret = 0
            context.stat.daycancels += ordernum

        if orderid in context.orders.keys():
            inst = context.orders[orderid]['instid']
            '''
            if context.orders[orderid]['action'] == 'open':
                if context.orders[orderid]['direction'] > 0:
                    context.longpendvol[inst]['volume'] = max(context.longpendvol[inst]['volume'] - context.orders[orderid]['leftvol'], 0)
                    if context.longpendvol[inst]['volume'] <= 0:
                        context.longpendvol[inst] = {'volume':0, 'poscost':0}
                else:
                    context.shortpendvol[inst]['volume'] = max(context.shortpendvol[inst]['volume'] - context.orders[orderid]['leftvol'], 0)
                    if context.shortpendvol[inst]['volume'] <= 0:
                        context.shortpendvol[inst] = {'volume':0, 'poscost':0}
            '''
            context.updateFrozenMarg()
            context.orders[orderid]['status'] = "canceled" if context.orders[orderid][
                                                                  'status'] != "parttraded" else "ptpc"
            # part traded part canceled
            context.orders[orderid]['cancelvol'] = context.orders[orderid]['leftvol']
            context.orders[orderid]['leftvol'] = 0
            if context.instid[-3:] == 'CCF':
                ## CCFX order commission
                context.stat.dayfees += 1
                context.stat.totalfees += 1
            ret = 0
            context.stat.daycancels += 1
        if ret == 0:
            context.stat.avail = context.stat.balance - context.marg - context.frozenMarg
            context.Cancels[context.curtime] = [orderid, context.stat.daycancels]
            # print(context.Cancels[context.curtime])
        return ret
    except Exception as e:
        logger.error(f"Failed on cancel order: {e.__traceback__.tb_lineno},{e}")
        return -1


def test_record_hedge_point(context, action, dirstr, volume):
    validaction_action = ['open', 'close']  # ['auto','open','close']

    if not action in validaction_action:
        print("action must be one of " + str(validaction_action) + ',your action:' + str(action))
        return

    validcation_dir = ['long', 'short']

    if not dirstr in validcation_dir:
        print('direction must be one of ' + str(validcation_dir) + ',your action' + str(dirstr))

    letters = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u',\
                 'v', 'w', 'x', 'y', 'z']
    mapDict = {}
    for i in range(len(context.instids)):
        inst = context.instids[i]
        mapDict[letters[i]] = context.dataslide[inst]['current'] if context.datamode == 'tick' else context.dataslide[inst]['close']
        
        
    current = eval(context.formula, mapDict)
    if  not context.curtime in context.d.index:
        context.d.loc[context.curtime, 'openlong'] = current if action=='open' and dirstr=='long' else np.nan
        context.d.loc[context.curtime, 'closelong'] = current if action=='close' and dirstr=='long' else np.nan
        context.d.loc[context.curtime, 'openshort'] = current if action=='open' and dirstr=='short' else np.nan
        context.d.loc[context.curtime, 'closeshort'] = current if action=='close' and dirstr=='short' else np.nan
    else:
        if action=='open' and dirstr=='long':
            context.d.loc[context.curtime, 'openlong'] = current  
        if action=='close' and dirstr=='long' :
            context.d.loc[context.curtime, 'closelong'] = current  
        if action=='open' and dirstr=='short' :
            context.d.loc[context.curtime, 'openshort'] = current 
        if action=='close' and dirstr=='short' :
            context.d.loc[context.curtime, 'closeshort'] = current  


def cancel_order(context, orderid):
    if context.runmode =='test':
        return test_cancel_order(context,orderid)
    else:
        return real_cancel_order(context,orderid)
        
        
def make_order(context, instid, direction, price, volume, ordertype="limit", action="open", closetype='auto', timecond='GFD', autoremake=[0,0,0,0], autocancel=[0,0], accid=-1,father=0):
    
    if context.runmode =='test':
        return test_make_order(context, instid, direction, price, volume, ordertype, action,closetype, timecond)
    else:
        return real_make_order(context, instid, direction, price, volume, ordertype,action,closetype,timecond, autoremake, autocancel, accid=accid,father=father)
        

def get_bar(context, freq, count=None):
    if context.runmode == 'test':
        return test_get_bar(context, freq, count)
    else:
       return real_get_bar(context,freq, count)
        


def test_get_bar(context, freq, count=None):
    if hasattr(context, 'bardata'):
        data = context.bardata
        if freq % context.freq != 0 or freq == 1:
            ### do not resample
            if count:
                for inst in data:
                    df = data[inst] 
                    if 'time' in df.columns:
                        del df['time']

                    data[inst]=df.dropna(how='all').iloc[-count:,]
            #del data['time']
            return data
        else:
            f=str(freq)+'min'
            
            for instid in data:
                try:
                    df=data[instid]
                    if len(df) > 0:
                        #print(df.time)
                        #df['runtime']= pd.to_datetime(df.time, format='%Y%m%d%H%M%S',errors='ignore')
                        #df.set_index(["runtime"], inplace=True)
   
                        df2 = pd.DataFrame(columns=df.columns)
                        #print('111',df2)
                        for col in df.columns:
                            tmp = pd.Series(index=df.index, data=df.loc[:,col])
                            if col == "open":
                                tmp=df[col].resample(f, label='right', closed='right').first()
                            elif col =="close":
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            elif col == 'high':
                                tmp=df[col].resample(f, label='right', closed='right').max()
                            elif col == 'low':
                                tmp=df[col].resample(f, label='right', closed='right').min()
                            elif col == 'volume':
                                tmp=df[col].resample(f, label='right', closed='right').sum()
                            elif col == 'money':
                                tmp=df[col].resample(f, label='right', closed='right').sum()
                            elif col == 'position':
                                tmp = df[col].resample(f, label='right', closed='right').last()
                            elif col == 'presett':
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            elif col == 'preclose':
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            elif col == 'lowerlimit':
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            elif col == 'upperlimit':
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            elif col == 'tradingday':
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            df2.loc[:,col] = tmp
                        #print(df2)
                        if 'time' in df2.columns:
                            del df2['time']
                        if isinstance(count, int):
                            data[instid]=df2.dropna(how='any').iloc[-count:,]
                        else:
                            data[instid]=df2.dropna(how='any')
                except Exception as e:
                    print("get_bar Error:", e.__traceback__.tb_lineno,e)
                
            return data
    else:
        return None

    

def real_get_bar(context, freq, count=None):
    data=get_bar_data(context)
    #print(data )
    if data:
        try:
            f=str(freq)+'min'
            for instid in data.keys():  
                df=data[instid]
                if len(df) > 0:
                    #print(df.time)
                    df['runtime']= pd.to_datetime(df.time, format='%Y%m%d%H%M%S',errors='ignore')
                    #for i in range(len(df['time'])):
                    #    df['runtime'].loc[i]=datetime.datetime.strptime(str(df['time'].loc[i]), "%Y%m%d%H%M%S")
                    df.set_index(["runtime"], inplace=True)
                    #print(df)
                    if f=='1min':
                        del data[instid]['time']
                        if count and isinstance(count, int):
                            
                            data[instid]=data[instid].iloc[-count:,]
                            
                    else:

                        df2 = pd.DataFrame(columns=df.columns)
                        #print('111',df2)
                        for col in df.columns:
                            tmp = pd.Series(index=df.index, data=df.loc[:,col])
                            if col == "open":
                                tmp=df[col].resample(f, label='right', closed='right').first()
                            elif col =="close":
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            elif col == 'high':
                                tmp=df[col].resample(f, label='right', closed='right').max()
                            elif col == 'low':
                                tmp=df[col].resample(f, label='right', closed='right').min()
                            elif col == 'volume':
                                tmp=df[col].resample(f, label='right', closed='right').sum()
                            elif col == 'money':
                                tmp=df[col].resample(f, label='right', closed='right').sum()
                            elif col == 'position':
                                tmp = df[col].resample(f, label='right', closed='right').last()
                            elif col == 'presett':
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            elif col == 'preclose':
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            elif col == 'lowerlimit':
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            elif col == 'upperlimit':
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            elif col == 'tradingday':
                                tmp=df[col].resample(f, label='right', closed='right').last()
                            df2.loc[:,col] = tmp
                        #print(df2)
                        del df2['time']
                        if isinstance(count, int):
                            data[instid]=df2.dropna(how='any').iloc[-count:,]
                        else:
                            data[instid]=df2.dropna(how='any')
        except Exception as e:
            print("get_bar Error:", e.__traceback__.tb_lineno,e)

        #print(df2.head())
    return data


def force_close(context):
    fcDict = context.fcDict
    for inst in fcDict:
        fclist = fcDict[inst]
        if fclist[0] > 0:
            make_order(context, inst, -1, context.getCurrent(inst),fclist[0],'market', 'close', 'closeyesterday')
        if fclist[1] > 0:
            make_order(context, inst, 1, context.getCurrent(inst),fclist[1],'market', 'close', 'closeyesterday')
        if fclist[2] > 0:    
            make_order(context, inst, -1, context.getCurrent(inst),fclist[2],'market', 'close', 'closetoday')
        if fclist[3] > 0:
            make_order(context, inst, 1, context.getCurrent(inst),fclist[3],'market', 'close', 'closetoday')
        




class testContext(object):
    instid = []
    def __init__(self):
        pass
        
def record_hedge_point(context, action, dirstr, volume):
    if context.runmode=='test':
        test_record_hedge_point(context, action, dirstr, volume)
    else:
        real_record_hedge_point(context, action, dirstr, volume)        
        

def reqCtpLogout(context,accid=0):
    from .qectpmarket_wrap import marketLogout 
    marketLogout()
    traderqueue = getAccidTraderQueue(accid)
    temporder = {}
    temporder['type'] = qetype.KEY_USER_LOGOUT
    traderqueue.put(temporder)
    
    


# -*- coding: utf-8 -*-
"""
Created on Sat Mar 19 13:07:37 2022

@author: ScottStation
"""

from .qeredisdb import saveAccountDatarealToDB, savePositionDatarealToDB,getDBAccountDetailReal
from .qeredisdb import delDBOrderDataReal, delDBTradeDataReal,getDBOrderDatareal
from datetime import datetime
from .qestatistics import g_stat
from .qeglobal import instSetts
from .qeriskctl import riskControl


class realAccountInfo:
    def __init__(self, balance=0,accid=0):
        self.balance = balance
        self.avail = balance
        self.margin = 0
        self.frozenMarg = 0
        self.pendVolumes = {}
        self.position = {}
        self.closeProf = 0
        self.posProf = 0
        self.daypnl = 0
        self.maxmarg = 0
        self.accupnl = 0
        self.totalpnl = 0
        self.dayfees = 0
        self.accufees = 0
        self.totalfees = 0
        self.withdraw = 0
        self.deposit = 0
        self.winamount = 0
        self.lossamount = 0
        self.wincount = 0
        self.losscount = 0
        self.turnover = 0
        self.current_timedigit = 0
        self.tradingDay = ""
        self.name='ctp'
        self.user = "unknown"
        self.investorid = ''
        ### internal parameters
        self.dataSlide = {}
        self.orders = {}
        self.trades = {}
        self.token = ''
        self.accid = accid
        self.riskctl = riskControl(self.riskctlCall)
        #self.g_order_id = int('3'+datetime.now().strftime('%Y%m%d')[2:])*100000
        self.g_trade_id = int('3'+datetime.now().strftime('%Y%m%d')[2:])*100000
        self.stgtable_load = False
        self.order_stg_table = {}
        self.loadReady = False

    def crossday(self):
        self.accupnl = self.totalpnl
        self.accufees = self.totalfees
        self.dayfees = 0
        self.daypnl = 0
        self.riskctl.crossDay()
        #self.g_order_id = int('3'+datetime.now().strftime('%Y%m%d')[2:])*100000
        self.g_trade_id = int('3'+datetime.now().strftime('%Y%m%d')[2:])*100000
        self.orders = {}
        self.trades = {}
        self.deposit = 0
        self.withdraw =  0
        self.turnover = 0
        self.maxmarg = 0
        self.winamount = 0
        self.lossamount = 0
        self.wincount = 0
        self.losscount = 0

        #self.pendVolumes = {}
        self.frozenMarg = 0
        for inst in self.position.keys():
            self.position[inst]['long']['yesvol'] = self.position[inst]['long']['volume']
            self.position[inst]['short']['yesvol'] = self.position[inst]['short']['volume']
        return

    def riskctlCall(self):
        instPrices = {}
        for oid in self.orders:
            order = self.orders[oid]
            if order['leftvol'] > 0 and order['ordertype'] == 'limit':
                field = 'longhigh' if order['direction'] > 0 else 'shortlow'
                instid = order['instid']
                price = order['price']
                if not instid in instPrices:
                    instPrices[instid] = {field:price}
                elif not field in  instPrices[instid]:
                    instPrices[instid][field] = price
                else:
                    if field == 'longhigh' and price > instPrices[instid][field]:
                        instPrices[instid][field] = price
                    elif field == 'shortlow' and price < instPrices[instid][field]:
                        instPrices[instid][field] = price     
        return instPrices                
    def calcFrozenVol(self, instid, direction):
        fvol = 0
        for oid in self.orders:
            order = self.orders[oid]
            if order['leftvol'] > 0 and order['action']=='close' and order['direction'] == direction:
                fvol += order['leftvol']
        return fvol
    
    def setLoadReady(self):
        self.loadReady = True


    def getPosProf(self):
        posProf = 0
        for inst in self.position.keys():
            current = self.dataSlide[inst]['current']
            volMult = instSetts[inst]['volmult']
            if self.position[inst]['long']['volume'] > 0:
                posProf += volMult * (current - self.position[inst]['long']['poscost']) * \
                           self.position[inst]['long']['volume']
            if self.position[inst]['short']['volume'] > 0:
                posProf += volMult * (self.position[inst]['short']['poscost'] - current) * \
                           self.position[inst]['short']['volume']
        return posProf 
    
    def getCommission(self,  action, price, vol, closetype,instid, instSetts):
        if action == 'open':

            return instSetts[instid]['openfee'] * vol + instSetts[instid]['openfeerate'] * vol * price * \
                   instSetts[instid]['volmult']
        else:
            rate = instSetts[instid]['closetodayrate'] if closetype == 'closetoday' else 1
            return  rate * (instSetts[instid]['closefee'] * vol + \
                                                          instSetts[instid]['closefeerate'] * vol * price * instSetts[instid]['volmult'])
    
    def getCloseProf(self, poscost, dirstr, price, vol, instid, instSetts):
        volMult = instSetts[instid]['volmult']
        if dirstr == 'short':
            closeProf = volMult * vol * (poscost - price)
        else:
            closeProf = volMult * vol * (price - poscost)
        return closeProf
    
    
    def getPosition(self,instid, dirstr, field):
        if instid in self.position:
            if dirstr in self.position[instid]:
                if field in self.position[instid][dirstr]:
                    return self.position[instid][dirstr][field]
        return 0        
    
    
    
    def updateWinLossParas(self, dirstr, price, vol, closetype, instid):
        poscost = self.getPosition(instid,dirstr,'poscost')
        tfee = self.getCommission('open', poscost, vol, closetype, instid, instSetts)
        tfee += self.getCommission('close', price, vol, closetype, instid, instSetts)
        tpnl = self.getCloseProf(poscost, dirstr, price, vol, instid, instSetts) - tfee
        if tpnl > 0:
            self.winamount += tpnl
            self.wincount += 1
        elif tpnl < 0:
            self.lossamount += tpnl
            self.losscount += 1
    
#     def updateMargin(self):
#         global g_dataSlide
#         marg = 0
#         for inst in self.position.keys():
#             current = g_dataSlide[inst]['current']
#             marg += self.getMargin(current, True, self.position[inst]['long']['volume'], inst)
#             marg += self.getMargin(current, False, self.position[inst]['short']['volume'], inst)
#         self.margin = marg
#         return    
#     def updateFrozenMarg(self):
#         global g_dataSlide
#         fm = 0
#         for inst in self.pendVolumes.keys():
#             current = g_dataSlide[inst]['current']
#             fm += self.getMargin(current, True, self.pendVolumes[inst]['long'], inst)
#             fm += self.getMargin(current, False, self.pendVolumes[inst]['short'], inst)
#         self.frozenMarg = fm
#         return
#     def updateCapital(self, closeProf, fee):
#         lastPosProf = self.posProf
#         self.posProf = self.getPosProf()
#         self.balance += closeProf - fee -lastPosProf + self.posProf
#         self.avail = self.balance -self.margin - self.frozenMarg
#         return
#     def getMargin(self, current, bLong, vol, instid):
#         marginrate = instSetts[instid]['marglong'] if bLong else instSetts[instid]['margshort']
#         return current * marginrate / 100 * vol * instSetts[instid]['volmult']
#     def getCommission(self,  action, price, vol, closetype,instid):
#         global instSetts   
#         if action == 'open':
#             return instSetts[instid]['openfee'] * vol + instSetts[instid]['openfeerate'] * vol * price * \
#                    instSetts[instid]['volmult']
#         else:
#             rate = instSetts[instid]['closetodayrate'] if closetype == 'closetoday' else 1
#             return  rate * (instSetts[instid]['closefee'] * vol + \
#                                                           instSetts[instid]['closefeerate'] * vol * price * instSetts[instid]['volmult'])
#         return
#     def readAndSetDB(self):
#         import math
#         try:
#             #load from db ,if not exist, write default value
#             bal =  getSimuInitCap(self.user)
#             if bal is None or math.isnan(float(bal)):
#                 updateInitCap(self.user, self.balance)
#             else:
#                 logger.info(f'load balance:{self.balance}')
#                 self.balance = float(bal)
#         except Exception as e:
#             logger.error(f"account.readAndSetDB error {e}",exc_info=True ) 
#         return
    def saveToDB(self):
        d = {}
        d['posProf'] = str(round(self.posProf,3))
        d['frozenMarg'] = str(round(self.frozenMarg,3))
        d['marg'] = str(round(self.margin,3))
        d['closeProf'] = str(round(self.closeProf,3))
        d['balance'] = str(round(self.balance,3))
        d['avail'] = str(round(self.avail,3))
        d['totalpnl'] = str(round(self.totalpnl,3))
        d['totalfee'] = str(round(self.totalfees,3))
        d['daypnl'] = str(round(self.daypnl,3))
        d['dayfee'] = str(round(self.dayfees,3))
        d['turnover'] = str(round(self.turnover,3))
        d['winamount'] = str(round(self.winamount,3))
        d['lossamount'] = str(round(self.lossamount,3))
        d['wincount'] = str(round(self.wincount,3))
        d['losscount'] = str(round(self.losscount,3))
        d['maxmarg'] = str(round(self.maxmarg,3))
       
        #print("Account save to DB")
        saveAccountDatarealToDB(self.user, self.investorid, self.tradingDay, d)
        savePositionDatarealToDB(self.user, self.investorid, self.position, self.tradingDay)
        g_stat.updateData(self.balance, self.daypnl, self.dayfees, self.margin, self.turnover, self.tradingDay,\
                          self.winamount, self.lossamount, self.wincount, self.losscount,self.maxmarg,\
                          accid=self.accid, withdraw=self.withdraw, deposit=self.deposit)
#         updateInitCap(self.user, self.balance)
#         logger.info(f'saveToDB {self.balance}')
    

    def loadFromDB(self,tradingday):
        res = getDBAccountDetailReal(self.user, self.investorid, str(tradingday))
        if res:
            d = eval(res)
            self.winamount = float(d['winamount'])
            self.lossamount = float(d['winamount'])
            self.wincount = float(d['wincount'])
            self.losscount = float(d['losscount'])
            self.maxmarg = float(d['maxmarg'])
        g_stat.loadFromDBReal(int(tradingday)) 
        orders = getDBOrderDatareal(self.user, self.investorid, tradingday)
        for oid in orders:
            stgname = eval(orders[oid])['stratName']
            self.order_stg_table[oid] = stgname
        #print(self.order_stg_table)
        self.stgtable_load = True  
        delDBOrderDataReal(self.user, self.investorid, tradingday)
        delDBTradeDataReal(self.user, self.investorid, tradingday)
    
    
#realaccount = realAccountInfo()




# -*- coding: utf-8 -*-
"""
Created on Sat Mar 12 16:02:44 2022

@author: ScottStation
"""

from .qeredisdb import SET, GET
import json
########################################################################
## Test cases
## 1. arbit_open-- ok
## 2. cancel by canceldiff -- ok
## 3. secondary orders --ok
## 4. force close by closediff --ok
## 5. chase order
## 6. close unfinished
## 7. secondary by 'component' 'quoted'
## 8. maxchase logic
## 9. force close fail and chase
#######################################################################


class qeArbitModel:
    
    arbit_orders={}
    arbit_order_id = 0
    
    def __init__(self, user, order_func, cancel_func,usedb=False):
        self.user = user
        self.order_func = order_func
        self.cancel_func = cancel_func
        self.bCrossDay = False
        self.fcstarted = False
        self.needsave = False
        self.usedb = usedb
        if usedb:
            self.loadFromDB()
        
    
    def arbit_open(self, context, action,  orderdict, destvol, canceldiff, closediff, chaseseconds,closeseconds=180, secondtype='market', todaypos = True, maxchase=2):
        '''

        Parameters
        ----------
        context : TYPE
            DESCRIPTION.
        action : str
            ['open','close'].
        orderdict : dict
            i.e. {'instid1':｛'price':1000, 'volbase': 1, 'dir': 1, 'tradefirst': True｝,'instid2':{'price':1200, 'volbase':1, 'dir': -1, 'tradefirst': False}}
            volbase: trade volume = destvol * volbase
            dir : Bid or Sale
            tradefirst: If this instid will be trade first, if it is true, the volbase must be 1. Only on instid could be traded first.
        destvol: int
            destination valume. it multiply the volbase is detination volume of each instid of orderdict
        canceldiff : float
            price difference to cancel order
        closediff : dict
            {'instid1':3.0, 'instid2':4.0}
            price difference to close order after first order traded, key is instrumentid
        chaseseconds: int
            if cannot trade in such seconds, make chase order
        closeseconds: int
            retry after closeseconds if not traded 
        todaypos: bool
            True: open or  close today False: close yesterday
        secondtype : str, optional
            'market' order or 'opponent'/'quoted'/'initial'/'current' order when trade second batch orders . The default is 'market'.
        maxchase : int, optional
            max number of chase orders. The default is 2.

        Returns
        -------
        None.

        '''
        sampledict ={'price':100.0, 'volbase':1, 'dir':1, 'tradefirst':False}
        validaction =['open','close']
        validtype =['market','opponent','quoted','initial','current']
        ## input paramters check
        if not action in validaction:
            print("Invalid aciton ,should be in",validaction)
            return -1
        
        if not action in validaction:
            print("Invalid order type ,should be in",validtype)
            return -1
        
        if not isinstance(closediff, dict):
            print("Invalide paramter type of closediff, should be float")
            return -1
           
        
        if  not isinstance(canceldiff, float) and not isinstance(canceldiff, int):
            print("Invalide paramter type of canceldiff, should be float or int")
            return -1
            
        if not isinstance(destvol, int) or not isinstance(maxchase,int)  or not isinstance(chaseseconds, int) or not isinstance(closeseconds, int):
            print("Invalid paramter type of destvol/maxchase/chaseseconds/closeseconds, should be int")
            return -1
        
        if not isinstance(todaypos, bool):
            print("Invalid paramter type of todaypos, should be Boolean")
            return -1
        
        for cd in closediff:
            if not cd in orderdict:
                print('key of closediff must be in orderdict')
                return -1
            elif not isinstance(closediff[cd], float) and not isinstance(closediff[cd], int):
                print('closediff values must be int or float numbers')
                return -1
        try:

            firstid = ''       
            for key in orderdict:
                if not key in context.instid:
                    print('The key of context should be instrumentid in context.instid')
                    return -1
                
                if orderdict[key]['tradefirst']:
                    if firstid == '':
                        firstid = key
                        if orderdict[key]['volbase'] != 1:
                            print('The volbase of instrumentid which be traded first should be 1 ')
                            return -1
                    else:
                        print("Only support one instid in orderdict.keys be traded first ")
                        return -1
                    
                for okey in sampledict:
                    if  not okey in orderdict[key].keys() or not isinstance(orderdict[key][okey], type(sampledict[okey])):
                        print("Please check the format of orderdict on key :",key, "example dict:", sampledict)
                        return -1
                ##Check the upperlimit and lowerlimit
                pdir = orderdict[key]['dir'] 
                if 'upperlimit' in  context.dataslide[key] and 'lowerlimit' in context.dataslide[key]:
                    if pdir > 0 and orderdict[key]['price'] == context.dataslide[key]['upperlimit']:
                        print("Price touch the upperlimit so do nothing.",key, orderdict[key]['price'])
                        return  -1
                    if pdir < 0 and orderdict[key]['price'] == context.dataslide[key]['lowerlimit']:
                        print("Price touch the lowerlimit so do nothing.",key, orderdict[key]['price'])
                        return -1
            
            ## make first order
            closetype = 'closeyesterday' if not todaypos else 'closetoday'
            #closetype ='none' if action =='open' else closetype
           
            pdir = orderdict[firstid]['dir']  
            vol = destvol * orderdict[firstid]['volbase']
            orderid = self.order_func(context, firstid, pdir, orderdict[firstid]['price'],vol,\
                            ordertype="limit", action=action, closetype = closetype)
            
            ## register the information
            if orderid > 0:
                self.arbit_order_id += 1
                    
                self.arbit_orders[self.arbit_order_id] = { 'orderdict': orderdict,
                    'destvol': destvol,                                      
                    'instids': list(orderdict.keys()),
                    'todaypos' : todaypos, 
                    'canceldiff': canceldiff,
                    'closediff': closediff,
                    'chaseseconds': chaseseconds,
                    'firstinst': firstid,
                    'maxchase': maxchase,
                    'status': 'unfinished',
                    'action': action,
                    'secondtype': secondtype,
                    'firstorderid': orderid,
                    'fordercancel': False,
                    'secondorders': [],
                    'closingorders': {},
                    'closeseconds':closeseconds,
                    'tradevol': 0,
                    'fclosevol': 0,
                    'cancelvol': 0,
                    }
                print( context.curtime, "arbit open",self.arbit_order_id)
                self.needsave = True
                return self.arbit_order_id
            else:
                return -1
        except Exception as e:
            print('arbit open error', e.__traceback__.tb_lineno,e)
            
            
            
            
    def force_close_unfinished(self,context):
        if self.fcstarted:
            return 
        
        self.fcstarted = True
        try:
            
            allfinished = True
            for key in self.arbit_orders.keys():
                forderid = self.arbit_orders[key]['firstorderid']
                fcanceled = self.arbit_orders[key]['fordercancel']
                if self.arbit_orders[key]["status"] != 'finished' :
                     allfinished = False
                     tmplist = []
                     if context.orders[forderid]['leftvol'] > 0 and not fcanceled:
                         self.cancel_func(context,forderid)
                         self.arbit_orders[key]['fordercancel'] = True
                     for so in self.arbit_orders[key]['secondorders']:
                          if so['status'] != 'finished' and so['status'] != 'forceclosing':
                              ## force close
                              so['status'] = 'foceclosing'      
                              for sokey in so['orderinfo']:
                                    if so['orderinfo'][sokey]['tradevol'] + so['orderinfo'][sokey]['cancelvol'] < so['orderinfo'][sokey]['destvol']:
                                            orderid = so['orderinfo']['orderid']
                                            self.cancel_func(context, orderid)
                          tmplist.append(so)
                          
                     self.arbit_orders[key]['secondorders']  = tmplist
            if not allfinished:
                print("force close unfinished")
        except Exception as e:
            print('arbit open error', e.__traceback__.tb_lineno,e)
                  
    
    def trace_order(self,context):
        '''
        

        Parameters
        ----------
        context : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        '''
        try:
            allfinished = True
            for key in self.arbit_orders.keys():
                if self.arbit_orders[key]["status"] != 'finished' :
                        ## check if it is traded
                        inst = self.arbit_orders[key]['firstinst']
                        forderid = self.arbit_orders[key]['firstorderid']
                        fcanceled = self.arbit_orders[key]['fordercancel']
                        tradedvol = self.arbit_orders[key]['tradevol']
                        orderdict = self.arbit_orders[key]['orderdict']
                        action = self.arbit_orders[key]['action']
                        todaypos = self.arbit_orders[key]['todaypos']
                        ordertype = self.arbit_orders[key]['secondtype']
                        #ordertype = 'limit' if not  =='market' else 'market'
                        closetype = 'closeyesterday' if not todaypos else 'closetoday'
                        closediff = self.arbit_orders[key]['closediff']
                        chaseseconds = self.arbit_orders[key]['chaseseconds']
                        closeseconds = self.arbit_orders[key]['closeseconds']
                        maxchase = self.arbit_orders[key]['maxchase']
                        
                        if  context.orders[forderid]['tradevol'] >  tradedvol:
                            ##更新优先合约交易成功vol
                            tradevol = context.orders[forderid]['tradevol'] - tradedvol
                            print(key, "first instid Traded ",tradevol)
                            self.arbit_orders[key]['tradevol'] = context.orders [forderid]['tradevol']
                            secondorders = {'firstvol': tradevol,'status':'unfinished','curtime':context.curtime, 'chasenum': 0}
                            orderinfo = {}
                            orderfail = False
                            for instid in self.arbit_orders[key]['orderdict']:
                                if instid != inst: 
                                    pdir = orderdict[instid]['dir']
                                    orderprice = orderdict[instid]['price']
                                    if ordertype == 'current':
                                        orderprice = context.current[instid]
                                    elif ordertype == 'opponent':
                                        orderprice = context.dataslide[instid]['al_p'] if pdir > 0 else  context.dataslide[instid]['bl_p']
                                    elif ordertype == 'quoted':
                                        orderprice = context.dataslide[instid]['bl_p'] if pdir > 0 else  context.dataslide[instid]['al_p']
                                    if ordertype != 'market':
                                        ordertype = 'limit'
                                    ordervol =     orderdict[instid]['volbase']*tradevol
                                    orderid = self.order_func(context, instid, pdir, orderprice, ordervol,\
                                                    ordertype=ordertype, action=action, closetype = closetype)
                                    if orderid > 0:
                                        orderinfo[instid] = {'orderid':orderid, 'destvol':ordervol, 'tradevol':0, 'cancelvol':0, 'dir':pdir,'chasing':False}    
                                    else:
                                        print(key, "Error to make secondary orders")
                                        orderfail = True
                                        break
                            secondorders['orderinfo'] = orderinfo
                            if orderfail:
                                for inst in orderinfo:
                                    self.cancel_func(context, orderinfo[instid]['orderid'])
                                if context.orders[forderid]['leftvol'] >0 and not fcanceled :
                                        self.cancel_func(context,forderid)
                                        self.arbit_orders[key]['fordercancel'] = True
        
                                secondorders['status'] = 'forceclosing'   
                            self.arbit_orders[key]['secondorders'].append( secondorders)
                                
    
     
                        elif context.current[inst] - self.arbit_orders[key]['orderdict'][inst]['price'] > self.arbit_orders[key]['canceldiff'] and not self.arbit_orders[key]['fordercancel'] :
                            ## 若满足canceldiff 取消剩余仓位
                            print(key, "cancel diff cancel")
                            self.cancel_func(context, forderid)
                            self.arbit_orders[key]['fordercancel'] = True
                        
                        tmplist = []    
                        #处理第二批订单
                        #secondTraded = True
                        for so in self.arbit_orders[key]['secondorders']:
                            if so['status'] != 'finished' and so['status'] != 'forceclosing':
                                ## check second orders.
                                for sokey in  so['orderinfo']:
                                         ## Check trade/chase/Close
                                         orderid = so['orderinfo'][sokey]['orderid']
                                         pdir = so['orderinfo'][sokey]['dir']
                                         if context.orders[orderid]['tradevol'] > so['orderinfo'][sokey]['tradevol']:
                                             ## 订单成交
                                             print(key, "Traded", sokey, context.orders[orderid]['tradevol'] - so['orderinfo'][sokey]['tradevol'])
                                             so['orderinfo'][sokey]['tradevol'] =  context.orders[orderid]['tradevol']
                                             
                                             
                                         elif  context.orders[orderid]['errorid'] != 0 or abs(context.current[sokey] - self.arbit_orders[key]['orderdict'][sokey]['price'])> closediff[sokey]:
                                             ## Force close
                                             so['status'] ='forceclosing'
                                             print(key, 'order error/closediff, forceclosing', sokey ,orderid, context.orders[orderid]['errorid'])
                                             if context.orders[orderid]['cancelvol'] > 0:
                                                 so['orderinfo'][sokey]['cancelvol'] = context.orders[orderid]['cancelvol']
                                             break
                                         elif (context.curtime - so['curtime']).seconds > chaseseconds:
                                             so['chasenum'] += 1
                                             if so['chasenum'] > maxchase:
                                                 ## 追单超过次数 Force close
                                                 print(key, "Exceed maxchase force closing")
                                                 so['status'] ='forceclosing'
                                                 break
                                             elif not so['orderinfo'][sokey]['chasing'] :
                                                 ## 触发追单
                                                 print(key, "chase order", sokey)
                                                 so['orderinfo'][sokey]['chasing'] = True
                                                 self.cancel_func(context, orderid)
                                                 pass                                        
                                            
                                         if context.orders[orderid]['cancelvol'] >  so['orderinfo'][sokey]['cancelvol']:
                                             so['orderinfo'][sokey]['cancelvol'] = context.orders[orderid]['cancelvol']
                                             if so['orderinfo'][sokey]['chasing']:
                                                 orderprice = context.dataslide[sokey]['al_p'] if pdir > 0 else  context.dataslide[sokey]['bl_p']
                                                 ordervol = so['orderinfo'][sokey]['cancelvol']
                                                 so['orderinfo'][sokey]['orderid'] = self.order_func(context, sokey, pdir, orderprice,ordervol,\
                                                    ordertype='limit', action=action, closetype = closetype)
                                                 so['orderinfo'][sokey]['cancelvol'] = 0  
                                                 so['curtime'] = context.curtime
                                                 so['orderinfo'][sokey]['destvol'] -= so['orderinfo'][sokey]['tradevol']
                                                 so['orderinfo'][sokey]['tradevol'] = 0
                                                 so['orderinfo'][sokey]['chasing'] = False
                                                 print(key, 'chase order', sokey, orderprice, ordervol)
     
                                if  so['status'] == 'forceclosing':   
                                    print(key, "forceclosing canceling")
                                    for sokey in so['orderinfo']:
                                        if so['orderinfo'][sokey]['tradevol'] + so['orderinfo'][sokey]['cancelvol'] < so['orderinfo'][sokey]['destvol']:
                                            orderid = so['orderinfo'][sokey]['orderid']
                                            self.cancel_func(context, orderid)
                                    if context.orders[forderid]['leftvol'] >0 and not fcanceled :
                                        self.cancel_func(context,forderid)
                                        self.arbit_orders[key]['fordercancel'] = True
                                    ##self.arbit_orders[key]['status'] = 'forceclosing'    
                                
                                alltraded = True
                                for sokey in so['orderinfo']:
                                    if so['orderinfo'][sokey]['tradevol'] < so['orderinfo'][sokey]['destvol']:
                                            alltraded = False
    
                                
                                if alltraded:
                                    #print(key, so['orderinfo'])
                                    so['status'] = 'finished'
                            elif so['status'] == 'forceclosing':
                                allcanceled = True
                                
                                for sokey in  so['orderinfo']:
                                    
                                    orderid = so['orderinfo'][sokey]['orderid']
                                    if context.orders[orderid]['leftvol'] >0:
                                        allcanceled = False
                                #print(allcanceled,so['orderinfo'][sokey] )
                                if allcanceled and  context.orders[forderid]['leftvol'] == 0:
                                    #self.arbit_orders[key]['status'] = 'forceclosing'
                                    so['status'] = 'finished'
                                    corders = {}
                                    if self.arbit_orders[key]['action'] =='open':
                                        ## force closing opened group
                                        print(key, 'Force closing left open position')
    
                                        firstvol = so['firstvol']
                                        self.arbit_orders[key]['fclosevol'] += firstvol
                                        pdir = - orderdict[inst]['dir']
                                        price = context.current[inst]
                                        orderid =  self.order_func(context, inst, pdir, price,firstvol,\
                                                    ordertype='market', action='close', closetype = closetype)
                                        corders[inst] = {'orderid':orderid, 'tradevol':0, 'destvol':firstvol,'dir':pdir,'curtime':context.curtime,'finished':False}
                                        for sokey in so['orderinfo']:
                                            vol = context.orders[so['orderinfo'][sokey]['orderid']]['tradevol']
                                            if vol > 0:
                                                price = context.current[sokey]
                                                pdir = - so['orderinfo'][sokey]['dir']
                                                orderid =  self.order_func(context, sokey, pdir, price, vol,\
                                                            ordertype='market', action='close', closetype = closetype)
                                                corders[sokey] = {'orderid':orderid, 'tradevol':0, 'destvol':vol,'dir':pdir,'curtime':context.curtime,'finished':False}
                                            
                                            
                                    elif self.arbit_orders[key]['action'] =='close':
                                        ## force closing letf group
                                        print(key, 'Force closing left close position')
                                        for sokey in so['orderinfo']:
                                            so['orderinfo'][sokey]['tradevol'] = context.orders[so['orderinfo'][sokey]['orderid']]['tradevol']
                                            vol = so['orderinfo'][sokey]['destvol'] - so['orderinfo'][sokey]['tradevol']
                                            if vol > 0:
                                                price = context.current[sokey]
                                                pdir =  so['orderinfo'][sokey]['dir']
                                                orderid =  self.order_func(context, sokey, pdir, price, vol,\
                                                            ordertype='market', action='close', closetype = closetype)
                                                corders[sokey] = {'orderid':orderid, 'tradevol':0, 'destvol':vol,'dir':pdir,'curtime':context.curtime,'finished':False}
                                    
                                    self.arbit_orders[key]['closingorders'] = corders
                                
                            tmplist.append(so)
                        self.arbit_orders[key]['secondorders'] = tmplist
                        
                        closenum = 0
                        
                        if len(self.arbit_orders[key]['closingorders']) > 0:
                            ## check trade responese
                            for instid in self.arbit_orders[key]['closingorders']:
                                corders = self.arbit_orders[key]['closingorders'][instid]
                                if not corders['finished']:
                                    orderid = corders['orderid']
                                    if context.orders[orderid]['tradevol'] > corders['tradevol']:
                                        ## New trade
                                        print(key, "force close traded ", instid, context.orders[orderid]['tradevol'] - corders['tradevol'])
                                        self.arbit_orders[key]['closingorders'][instid]['tradevol'] = context.orders[orderid]['tradevol']
                                        if corders['tradevol'] == corders['destvol']:
                                            self.arbit_orders[key]['closingorders'][instid]['finished'] = True
                                            
                                            closenum += 1
                                    if context.orders[orderid]['errorid'] != 0 and (context.curtime - corders['curtime']).seconds > closeseconds:
                                        ## chase order
                                        vol = corders['destvol'] - corders['tradevol']
                                        print(key, "force closing chase order", vol)
                                        orderid = self.order_func(context, instid, corders['dir'], context.current[instid], vol,\
                                                            ordertype='market', action='close', closetype = closetype)
                                        corders['orderid'] = orderid
                                        corders['curtime'] = context.curtime
                                        corders['destvol'] -= corders['tradevol']
                                        corders['tradevol'] = 0
                                        self.arbit_orders[key]['closingorders'][instid] = corders
                                            
                                else:
                                    closenum += 1
                              
                            
                        if  context.orders[forderid]['cancelvol'] > self.arbit_orders[key]['cancelvol'] :
                            ## 更新cancelvol
                            print(key, 'first contract canceled:', context.orders[forderid]['cancelvol'] - self.arbit_orders[key]['cancelvol'])
                            self.arbit_orders[key]['cancelvol'] = context.orders[forderid]['cancelvol']
                            
                            
                        ## 判断是否全部完成
                        secondTraded = True
                        for so in self.arbit_orders[key]['secondorders']:
                            if not so['status'] =='finished':
                                secondTraded = False
                        if secondTraded and closenum == len(self.arbit_orders[key]['closingorders']) and  context.orders[forderid]['leftvol'] == 0:
                            
                            self.arbit_orders[key]["status"] = 'finished'
                            print(key, 'finished')
                        else:
                            allfinished = False
            
            if allfinished and self.needsave and self.usedb:
                self.saveToDB()
            return allfinished
        except Exception as e:
            print('arbit trace error', e.__traceback__.tb_lineno,e)
    
    def saveToDB(self):
        if len(self.arbit_orders) > 0:
            SET(self.user, 'ARBIT_STRUCT',json.dumps(self.arbit_orders))
    
    def loadFromDB(self):
        ostr = GET(self.user, 'ARBIT_STRUCT')
        if ostr:
            self.arbit_orders = json.loads(ostr)
    
            
    def cross_day(self,context):
        self.bCrossDay = True
        self.fcstarted = False
        print('arbit order crossday')
        for key in self.arbit_orders:
            self.arbit_orders[key]['todaypos'] = False
            self.arbit_orders[key]['status'] = 'finished'
            
            
            